@@ -7,7 +7,6 @@ import (
 	"encoding/json"
 	"fmt"
 	"io"
-	"os"
 
 	"github.com/spf13/cobra"
 
@@ -15,8 +14,6 @@ import (
 	"github.com/restic/restic/internal/pack"
 	"github.com/restic/restic/internal/repository"
 	"github.com/restic/restic/internal/restic"
-
-	"github.com/restic/restic/internal/worker"
 )
 
 var cmdDebug = &cobra.Command{
@@ -29,7 +26,13 @@ var cmdDebugDump = &cobra.Command{
 	Short: "Dump data structures",
 	Long: `
 The "dump" command dumps data structures from the repository as JSON objects. It
-is used for debugging purposes only.`,
+is used for debugging purposes only.
+
+EXIT STATUS
+===========
+
+Exit status is 0 if the command was successful, and non-zero if there was any error.
+`,
 	DisableAutoGenTag: true,
 	RunE: func(cmd *cobra.Command, args []string) error {
 		return runDebugDump(globalOptions, args)
@@ -51,27 +54,19 @@ func prettyPrintJSON(wr io.Writer, item interface{}) error {
 	return err
 }
 
-func debugPrintSnapshots(repo *repository.Repository, wr io.Writer) error {
-	for id := range repo.List(context.TODO(), restic.SnapshotFile) {
-		snapshot, err := restic.LoadSnapshot(context.TODO(), repo, id)
+func debugPrintSnapshots(ctx context.Context, repo *repository.Repository, wr io.Writer) error {
+	return repo.List(ctx, restic.SnapshotFile, func(id restic.ID, size int64) error {
+		snapshot, err := restic.LoadSnapshot(ctx, repo, id)
 		if err != nil {
-			fmt.Fprintf(os.Stderr, "LoadSnapshot(%v): %v", id.Str(), err)
-			continue
+			return err
 		}
 
 		fmt.Fprintf(wr, "snapshot_id: %v\n", id)
 
-		err = prettyPrintJSON(wr, snapshot)
-		if err != nil {
-			return err
-		}
-	}
-
-	return nil
+		return prettyPrintJSON(wr, snapshot)
+	})
 }
 
-const dumpPackWorkers = 10
-
 // Pack is the struct used in printPacks.
 type Pack struct {
 	Name string `json:"name"`
@@ -87,50 +82,22 @@ type Blob struct {
 	Offset uint            `json:"offset"`
 }
 
-func printPacks(repo *repository.Repository, wr io.Writer) error {
-	f := func(ctx context.Context, job worker.Job) (interface{}, error) {
-		name := job.Data.(string)
+func printPacks(ctx context.Context, repo *repository.Repository, wr io.Writer) error {
 
-		h := restic.Handle{Type: restic.DataFile, Name: name}
+	return repo.List(ctx, restic.PackFile, func(id restic.ID, size int64) error {
+		h := restic.Handle{Type: restic.PackFile, Name: id.String()}
 
-		blobInfo, err := repo.Backend().Stat(ctx, h)
+		blobs, err := pack.List(repo.Key(), restic.ReaderAt(ctx, repo.Backend(), h), size)
 		if err != nil {
-			return nil, err
+			Warnf("error for pack %v: %v\n", id.Str(), err)
+			return nil
 		}
 
-		blobs, err := pack.List(repo.Key(), restic.ReaderAt(repo.Backend(), h), blobInfo.Size)
-		if err != nil {
-			return nil, err
-		}
-
-		return blobs, nil
-	}
-
-	jobCh := make(chan worker.Job)
-	resCh := make(chan worker.Job)
-	wp := worker.New(context.TODO(), dumpPackWorkers, f, jobCh, resCh)
-
-	go func() {
-		for name := range repo.Backend().List(context.TODO(), restic.DataFile) {
-			jobCh <- worker.Job{Data: name}
-		}
-		close(jobCh)
-	}()
-
-	for job := range resCh {
-		name := job.Data.(string)
-
-		if job.Error != nil {
-			fmt.Fprintf(os.Stderr, "error for pack %v: %v\n", name, job.Error)
-			continue
-		}
-
-		entries := job.Result.([]restic.Blob)
 		p := Pack{
-			Name:  name,
-			Blobs: make([]Blob, len(entries)),
+			Name:  id.String(),
+			Blobs: make([]Blob, len(blobs)),
 		}
-		for i, blob := range entries {
+		for i, blob := range blobs {
 			p.Blobs[i] = Blob{
 				Type:   blob.Type,
 				Length: blob.Length,
@@ -139,30 +106,21 @@ func printPacks(repo *repository.Repository, wr io.Writer) error {
 			}
 		}
 
-		prettyPrintJSON(os.Stdout, p)
-	}
-
-	wp.Wait()
-
-	return nil
+		return prettyPrintJSON(wr, p)
+	})
 }
 
-func dumpIndexes(repo restic.Repository) error {
-	for id := range repo.List(context.TODO(), restic.IndexFile) {
-		fmt.Printf("index_id: %v\n", id)
+func dumpIndexes(ctx context.Context, repo restic.Repository, wr io.Writer) error {
+	return repo.List(ctx, restic.IndexFile, func(id restic.ID, size int64) error {
+		Printf("index_id: %v\n", id)
 
-		idx, err := repository.LoadIndex(context.TODO(), repo, id)
+		idx, err := repository.LoadIndex(ctx, repo, id)
 		if err != nil {
 			return err
 		}
 
-		err = idx.Dump(os.Stdout)
-		if err != nil {
-			return err
-		}
-	}
-
-	return nil
+		return idx.Dump(wr)
+	})
 }
 
 func runDebugDump(gopts GlobalOptions, args []string) error {
@@ -176,36 +134,31 @@ func runDebugDump(gopts GlobalOptions, args []string) error {
 	}
 
 	if !gopts.NoLock {
-		lock, err := lockRepo(repo)
+		lock, err := lockRepo(gopts.ctx, repo)
 		defer unlockRepo(lock)
 		if err != nil {
 			return err
 		}
 	}
 
-	err = repo.LoadIndex(gopts.ctx)
-	if err != nil {
-		return err
-	}
-
 	tpe := args[0]
 
 	switch tpe {
 	case "indexes":
-		return dumpIndexes(repo)
+		return dumpIndexes(gopts.ctx, repo, gopts.stdout)
 	case "snapshots":
-		return debugPrintSnapshots(repo, os.Stdout)
+		return debugPrintSnapshots(gopts.ctx, repo, gopts.stdout)
 	case "packs":
-		return printPacks(repo, os.Stdout)
+		return printPacks(gopts.ctx, repo, gopts.stdout)
 	case "all":
-		fmt.Printf("snapshots:\n")
-		err := debugPrintSnapshots(repo, os.Stdout)
+		Printf("snapshots:\n")
+		err := debugPrintSnapshots(gopts.ctx, repo, gopts.stdout)
 		if err != nil {
 			return err
 		}
 
-		fmt.Printf("\nindexes:\n")
-		err = dumpIndexes(repo)
+		Printf("\nindexes:\n")
+		err = dumpIndexes(gopts.ctx, repo, gopts.stdout)
 		if err != nil {
 			return err
 		}
