@@ -14,18 +14,23 @@ import (
 )
 
 var cmdDiff = &cobra.Command{
-	Use:   "diff snapshot-ID snapshot-ID",
+	Use:   "diff [flags] snapshot-ID snapshot-ID",
 	Short: "Show differences between two snapshots",
 	Long: `
 The "diff" command shows differences from the first to the second snapshot. The
 first characters in each line display what has happened to a particular file or
 directory:
 
- +  The item was added
- -  The item was removed
- U  The metadata (access mode, timestamps, ...) for the item was updated
- M  The file's content was modified
- T  The type was changed, e.g. a file was made a symlink
+* +  The item was added
+* -  The item was removed
+* U  The metadata (access mode, timestamps, ...) for the item was updated
+* M  The file's content was modified
+* T  The type was changed, e.g. a file was made a symlink
+
+EXIT STATUS
+===========
+
+Exit status is 0 if the command was successful, and non-zero if there was any error.
 `,
 	DisableAutoGenTag: true,
 	RunE: func(cmd *cobra.Command, args []string) error {
@@ -48,7 +53,7 @@ func init() {
 }
 
 func loadSnapshot(ctx context.Context, repo *repository.Repository, desc string) (*restic.Snapshot, error) {
-	id, err := restic.FindSnapshot(repo, desc)
+	id, err := restic.FindSnapshot(ctx, repo, desc)
 	if err != nil {
 		return nil, err
 	}
@@ -66,7 +71,7 @@ type Comparer struct {
 type DiffStat struct {
 	Files, Dirs, Others  int
 	DataBlobs, TreeBlobs int
-	Bytes                int
+	Bytes                uint64
 }
 
 // Add adds stats information for node to s.
@@ -111,10 +116,10 @@ func addBlobs(bs restic.BlobSet, node *restic.Node) {
 
 // DiffStats collects the differences between two snapshots.
 type DiffStats struct {
-	ChangedFiles            int
-	Added                   DiffStat
-	Removed                 DiffStat
-	BlobsBefore, BlobsAfter restic.BlobSet
+	ChangedFiles                         int
+	Added                                DiffStat
+	Removed                              DiffStat
+	BlobsBefore, BlobsAfter, BlobsCommon restic.BlobSet
 }
 
 // NewDiffStats creates new stats for a diff run.
@@ -122,6 +127,7 @@ func NewDiffStats() *DiffStats {
 	return &DiffStats{
 		BlobsBefore: restic.NewBlobSet(),
 		BlobsAfter:  restic.NewBlobSet(),
+		BlobsCommon: restic.NewBlobSet(),
 	}
 }
 
@@ -135,13 +141,13 @@ func updateBlobs(repo restic.Repository, blobs restic.BlobSet, stats *DiffStat)
 			stats.TreeBlobs++
 		}
 
-		size, err := repo.LookupBlobSize(h.ID, h.Type)
-		if err != nil {
-			Warnf("unable to find blob size for %v: %v\n", h, err)
+		size, found := repo.LookupBlobSize(h.ID, h.Type)
+		if !found {
+			Warnf("unable to find blob size for %v\n", h)
 			continue
 		}
 
-		stats.Bytes += int(size)
+		stats.Bytes += uint64(size)
 	}
 }
 
@@ -172,6 +178,27 @@ func (c *Comparer) printDir(ctx context.Context, mode string, stats *DiffStat, b
 	return nil
 }
 
+func (c *Comparer) collectDir(ctx context.Context, blobs restic.BlobSet, id restic.ID) error {
+	debug.Log("print tree %v", id)
+	tree, err := c.repo.LoadTree(ctx, id)
+	if err != nil {
+		return err
+	}
+
+	for _, node := range tree.Nodes {
+		addBlobs(blobs, node)
+
+		if node.Type == "dir" {
+			err := c.collectDir(ctx, blobs, *node.Subtree)
+			if err != nil {
+				Warnf("error: %v\n", err)
+			}
+		}
+	}
+
+	return nil
+}
+
 func uniqueNodeNames(tree1, tree2 *restic.Tree) (tree1Nodes, tree2Nodes map[string]*restic.Node, uniqueNames []string) {
 	names := make(map[string]struct{})
 	tree1Nodes = make(map[string]*restic.Node)
@@ -191,7 +218,7 @@ func uniqueNodeNames(tree1, tree2 *restic.Tree) (tree1Nodes, tree2Nodes map[stri
 		uniqueNames = append(uniqueNames, name)
 	}
 
-	sort.Sort(sort.StringSlice(uniqueNames))
+	sort.Strings(uniqueNames)
 	return tree1Nodes, tree2Nodes, uniqueNames
 }
 
@@ -243,7 +270,12 @@ func (c *Comparer) diffTree(ctx context.Context, stats *DiffStats, prefix string
 			}
 
 			if node1.Type == "dir" && node2.Type == "dir" {
-				err := c.diffTree(ctx, stats, name, *node1.Subtree, *node2.Subtree)
+				var err error
+				if (*node1.Subtree).Equal(*node2.Subtree) {
+					err = c.collectDir(ctx, stats.BlobsCommon, *node1.Subtree)
+				} else {
+					err = c.diffTree(ctx, stats, name, *node1.Subtree, *node2.Subtree)
+				}
 				if err != nil {
 					Warnf("error: %v\n", err)
 				}
@@ -300,7 +332,7 @@ func runDiff(opts DiffOptions, gopts GlobalOptions, args []string) error {
 	}
 
 	if !gopts.NoLock {
-		lock, err := lockRepo(repo)
+		lock, err := lockRepo(ctx, repo)
 		defer unlockRepo(lock)
 		if err != nil {
 			return err
@@ -340,8 +372,8 @@ func runDiff(opts DiffOptions, gopts GlobalOptions, args []string) error {
 	}
 
 	both := stats.BlobsBefore.Intersect(stats.BlobsAfter)
-	updateBlobs(repo, stats.BlobsBefore.Sub(both), &stats.Removed)
-	updateBlobs(repo, stats.BlobsAfter.Sub(both), &stats.Added)
+	updateBlobs(repo, stats.BlobsBefore.Sub(both).Sub(stats.BlobsCommon), &stats.Removed)
+	updateBlobs(repo, stats.BlobsAfter.Sub(both).Sub(stats.BlobsCommon), &stats.Added)
 
 	Printf("\n")
 	Printf("Files:       %5d new, %5d removed, %5d changed\n", stats.Added.Files, stats.Removed.Files, stats.ChangedFiles)
