@@ -2,7 +2,9 @@ package azure
 
 import (
 	"context"
+	"encoding/base64"
 	"io"
+	"io/ioutil"
 	"net/http"
 	"os"
 	"path"
@@ -63,13 +65,13 @@ func open(cfg Config, rt http.RoundTripper) (*Backend, error) {
 }
 
 // Open opens the Azure backend at specified container.
-func Open(cfg Config, rt http.RoundTripper) (restic.Backend, error) {
+func Open(cfg Config, rt http.RoundTripper) (*Backend, error) {
 	return open(cfg, rt)
 }
 
 // Create opens the Azure backend at specified container and creates the container if
 // it does not exist yet.
-func Create(cfg Config, rt http.RoundTripper) (restic.Backend, error) {
+func Create(cfg Config, rt http.RoundTripper) (*Backend, error) {
 	be, err := open(cfg, rt)
 
 	if err != nil {
@@ -114,19 +116,8 @@ func (be *Backend) Path() string {
 	return be.prefix
 }
 
-// preventCloser wraps an io.Reader to run a function instead of the original Close() function.
-type preventCloser struct {
-	io.Reader
-	f func()
-}
-
-func (wr preventCloser) Close() error {
-	wr.f()
-	return nil
-}
-
 // Save stores data in the backend at the handle.
-func (be *Backend) Save(ctx context.Context, h restic.Handle, rd io.Reader) (err error) {
+func (be *Backend) Save(ctx context.Context, h restic.Handle, rd restic.RewindReader) error {
 	if err := h.Valid(); err != nil {
 		return err
 	}
@@ -135,30 +126,22 @@ func (be *Backend) Save(ctx context.Context, h restic.Handle, rd io.Reader) (err
 
 	debug.Log("Save %v at %v", h, objName)
 
-	// Check key does not already exist
-	found, err := be.container.GetBlobReference(objName).Exists()
-	if err != nil {
-		return errors.Wrap(err, "GetBlobReference().Exists()")
-	}
-	if found {
-		debug.Log("%v already exists", h)
-		return errors.New("key already exists")
-	}
-
 	be.sem.GetToken()
 
-	// wrap the reader so that net/http client cannot close the reader, return
-	// the token instead.
-	rd = preventCloser{
-		Reader: rd,
-		f: func() {
-			debug.Log("Close()")
-		},
-	}
-
 	debug.Log("InsertObject(%v, %v)", be.container.Name, objName)
 
-	err = be.container.GetBlobReference(objName).CreateBlockBlobFromReader(rd, nil)
+	var err error
+	if rd.Length() < 256*1024*1024 {
+		// wrap the reader so that net/http client cannot close the reader
+		dataReader := ioutil.NopCloser(rd)
+
+		// if it's smaller than 256miB, then just create the file directly from the reader
+		err = be.container.GetBlobReference(objName).CreateBlockBlobFromReader(dataReader, nil)
+	} else {
+		// otherwise use the more complicated method
+		err = be.saveLarge(ctx, objName, rd)
+
+	}
 
 	be.sem.ReleaseToken()
 	debug.Log("%v, err %#v", objName, err)
@@ -166,6 +149,55 @@ func (be *Backend) Save(ctx context.Context, h restic.Handle, rd io.Reader) (err
 	return errors.Wrap(err, "CreateBlockBlobFromReader")
 }
 
+func (be *Backend) saveLarge(ctx context.Context, objName string, rd restic.RewindReader) error {
+	// create the file on the server
+	file := be.container.GetBlobReference(objName)
+	err := file.CreateBlockBlob(nil)
+	if err != nil {
+		return errors.Wrap(err, "CreateBlockBlob")
+	}
+
+	// read the data, in 100 MiB chunks
+	buf := make([]byte, 100*1024*1024)
+	var blocks []storage.Block
+
+	for {
+		n, err := io.ReadFull(rd, buf)
+		if err == io.ErrUnexpectedEOF {
+			err = nil
+		}
+		if err == io.EOF {
+			// end of file reached, no bytes have been read at all
+			break
+		}
+
+		if err != nil {
+			return errors.Wrap(err, "ReadFull")
+		}
+
+		buf = buf[:n]
+
+		// upload it as a new "block", use the base64 hash for the ID
+		h := restic.Hash(buf)
+		id := base64.StdEncoding.EncodeToString(h[:])
+		debug.Log("PutBlock %v with %d bytes", id, len(buf))
+		err = file.PutBlock(id, buf, nil)
+		if err != nil {
+			return errors.Wrap(err, "PutBlock")
+		}
+
+		blocks = append(blocks, storage.Block{
+			ID:     id,
+			Status: "Uncommitted",
+		})
+	}
+
+	debug.Log("uploaded %d parts: %v", len(blocks), blocks)
+	err = file.PutBlockList(blocks, nil)
+	debug.Log("PutBlockList returned %v", err)
+	return errors.Wrap(err, "PutBlockList")
+}
+
 // wrapReader wraps an io.ReadCloser to run an additional function on Close.
 type wrapReader struct {
 	io.ReadCloser
@@ -178,10 +210,13 @@ func (wr wrapReader) Close() error {
 	return err
 }
 
-// Load returns a reader that yields the contents of the file at h at the
-// given offset. If length is nonzero, only a portion of the file is
-// returned. rd must be closed after use.
-func (be *Backend) Load(ctx context.Context, h restic.Handle, length int, offset int64) (io.ReadCloser, error) {
+// Load runs fn with a reader that yields the contents of the file at h at the
+// given offset.
+func (be *Backend) Load(ctx context.Context, h restic.Handle, length int, offset int64, fn func(rd io.Reader) error) error {
+	return backend.DefaultLoad(ctx, h, length, offset, be.openReader, fn)
+}
+
+func (be *Backend) openReader(ctx context.Context, h restic.Handle, length int, offset int64) (io.ReadCloser, error) {
 	debug.Log("Load %v, length %v, offset %v from %v", h, length, offset, be.Filename(h))
 	if err := h.Valid(); err != nil {
 		return nil, err
@@ -242,7 +277,11 @@ func (be *Backend) Stat(ctx context.Context, h restic.Handle) (restic.FileInfo,
 		return restic.FileInfo{}, errors.Wrap(err, "blob.GetProperties")
 	}
 
-	return restic.FileInfo{Size: int64(blob.Properties.ContentLength)}, nil
+	fi := restic.FileInfo{
+		Size: int64(blob.Properties.ContentLength),
+		Name: h.Name,
+	}
+	return fi, nil
 }
 
 // Test returns true if a blob of the given type and name exists in the backend.
@@ -271,17 +310,15 @@ func (be *Backend) Remove(ctx context.Context, h restic.Handle) error {
 	return errors.Wrap(err, "client.RemoveObject")
 }
 
-// List returns a channel that yields all names of blobs of type t. A
-// goroutine is started for this. If the channel done is closed, sending
-// stops.
-func (be *Backend) List(ctx context.Context, t restic.FileType) <-chan string {
+// List runs fn for each file in the backend which has the type t. When an
+// error occurs (or fn returns an error), List stops and returns it.
+func (be *Backend) List(ctx context.Context, t restic.FileType, fn func(restic.FileInfo) error) error {
 	debug.Log("listing %v", t)
-	ch := make(chan string)
 
 	prefix, _ := be.Basedir(t)
 
 	// make sure prefix ends with a slash
-	if prefix[len(prefix)-1] != '/' {
+	if !strings.HasSuffix(prefix, "/") {
 		prefix += "/"
 	}
 
@@ -290,59 +327,63 @@ func (be *Backend) List(ctx context.Context, t restic.FileType) <-chan string {
 		Prefix:     prefix,
 	}
 
-	go func() {
-		defer close(ch)
+	for {
+		be.sem.GetToken()
+		obj, err := be.container.ListBlobs(params)
+		be.sem.ReleaseToken()
 
-		for {
-			be.sem.GetToken()
-			obj, err := be.container.ListBlobs(params)
-			be.sem.ReleaseToken()
-
-			if err != nil {
-				return
-			}
-
-			debug.Log("got %v objects", len(obj.Blobs))
-
-			for _, item := range obj.Blobs {
-				m := strings.TrimPrefix(item.Name, prefix)
-				if m == "" {
-					continue
-				}
-
-				select {
-				case ch <- path.Base(m):
-				case <-ctx.Done():
-					return
-				}
-			}
-
-			if obj.NextMarker == "" {
-				break
-			}
-			params.Marker = obj.NextMarker
+		if err != nil {
+			return err
 		}
-	}()
 
-	return ch
+		debug.Log("got %v objects", len(obj.Blobs))
+
+		for _, item := range obj.Blobs {
+			m := strings.TrimPrefix(item.Name, prefix)
+			if m == "" {
+				continue
+			}
+
+			fi := restic.FileInfo{
+				Name: path.Base(m),
+				Size: item.Properties.ContentLength,
+			}
+
+			if ctx.Err() != nil {
+				return ctx.Err()
+			}
+
+			err := fn(fi)
+			if err != nil {
+				return err
+			}
+
+			if ctx.Err() != nil {
+				return ctx.Err()
+			}
+
+		}
+
+		if obj.NextMarker == "" {
+			break
+		}
+		params.Marker = obj.NextMarker
+	}
+
+	return ctx.Err()
 }
 
 // Remove keys for a specified backend type.
 func (be *Backend) removeKeys(ctx context.Context, t restic.FileType) error {
-	for key := range be.List(ctx, restic.DataFile) {
-		err := be.Remove(ctx, restic.Handle{Type: restic.DataFile, Name: key})
-		if err != nil {
-			return err
-		}
-	}
-
-	return nil
+	return be.List(ctx, t, func(fi restic.FileInfo) error {
+		return be.Remove(ctx, restic.Handle{Type: t, Name: fi.Name})
+	})
 }
 
 // Delete removes all restic keys in the bucket. It will not remove the bucket itself.
 func (be *Backend) Delete(ctx context.Context) error {
 	alltypes := []restic.FileType{
-		restic.DataFile,
+		restic.PackFile,
 		restic.KeyFile,
 		restic.LockFile,
 		restic.SnapshotFile,
