@@ -3,22 +3,23 @@ package gs
 
 import (
 	"context"
-	"fmt"
 	"io"
+	"net/http"
 	"os"
 	"path"
 	"strings"
 
+	"cloud.google.com/go/storage"
 	"github.com/pkg/errors"
 	"github.com/restic/restic/internal/backend"
 	"github.com/restic/restic/internal/debug"
 	"github.com/restic/restic/internal/restic"
 
-	"io/ioutil"
-
+	"golang.org/x/oauth2"
 	"golang.org/x/oauth2/google"
 	"google.golang.org/api/googleapi"
-	storage "google.golang.org/api/storage/v1"
+	"google.golang.org/api/iterator"
+	"google.golang.org/api/option"
 )
 
 // Backend stores data in a GCS bucket.
@@ -29,10 +30,11 @@ import (
 //  * storage.objects.get
 //  * storage.objects.list
 type Backend struct {
-	service      *storage.Service
+	gcsClient    *storage.Client
 	projectID    string
 	sem          *backend.Semaphore
 	bucketName   string
+	bucket       *storage.BucketHandle
 	prefix       string
 	listMaxItems int
 	backend.Layout
@@ -41,36 +43,55 @@ type Backend struct {
 // Ensure that *Backend implements restic.Backend.
 var _ restic.Backend = &Backend{}
 
-func getStorageService(jsonKeyPath string) (*storage.Service, error) {
-
-	raw, err := ioutil.ReadFile(jsonKeyPath)
-	if err != nil {
-		return nil, errors.Wrap(err, "ReadFile")
+func getStorageClient(rt http.RoundTripper) (*storage.Client, error) {
+	// create a new HTTP client
+	httpClient := &http.Client{
+		Transport: rt,
 	}
 
-	conf, err := google.JWTConfigFromJSON(raw, storage.DevstorageReadWriteScope)
+	// create a new context with the HTTP client stored at the oauth2.HTTPClient key
+	ctx := context.WithValue(context.Background(), oauth2.HTTPClient, httpClient)
+
+	var ts oauth2.TokenSource
+	if token := os.Getenv("GOOGLE_ACCESS_TOKEN"); token != "" {
+		ts = oauth2.StaticTokenSource(&oauth2.Token{
+			AccessToken: token,
+			TokenType:   "Bearer",
+		})
+	} else {
+		var err error
+		ts, err = google.DefaultTokenSource(ctx, storage.ScopeReadWrite)
+		if err != nil {
+			return nil, err
+		}
+	}
+
+	oauthClient := oauth2.NewClient(ctx, ts)
+
+	gcsClient, err := storage.NewClient(ctx, option.WithHTTPClient(oauthClient))
 	if err != nil {
 		return nil, err
 	}
 
-	client := conf.Client(context.TODO())
+	return gcsClient, nil
+}
 
-	service, err := storage.New(client)
-	if err != nil {
-		return nil, err
+func (be *Backend) bucketExists(ctx context.Context, bucket *storage.BucketHandle) (bool, error) {
+	_, err := bucket.Attrs(ctx)
+	if err == storage.ErrBucketNotExist {
+		return false, nil
 	}
-
-	return service, nil
+	return err == nil, err
 }
 
 const defaultListMaxItems = 1000
 
-func open(cfg Config) (*Backend, error) {
+func open(cfg Config, rt http.RoundTripper) (*Backend, error) {
 	debug.Log("open, config %#v", cfg)
 
-	service, err := getStorageService(cfg.JSONKeyPath)
+	gcsClient, err := getStorageClient(rt)
 	if err != nil {
-		return nil, errors.Wrap(err, "getStorageService")
+		return nil, errors.Wrap(err, "getStorageClient")
 	}
 
 	sem, err := backend.NewSemaphore(cfg.Connections)
@@ -79,10 +100,11 @@ func open(cfg Config) (*Backend, error) {
 	}
 
 	be := &Backend{
-		service:    service,
+		gcsClient:  gcsClient,
 		projectID:  cfg.ProjectID,
 		sem:        sem,
 		bucketName: cfg.Bucket,
+		bucket:     gcsClient.Bucket(cfg.Bucket),
 		prefix:     cfg.Prefix,
 		Layout: &backend.DefaultLayout{
 			Path: cfg.Prefix,
@@ -95,8 +117,8 @@ func open(cfg Config) (*Backend, error) {
 }
 
 // Open opens the gs backend at the specified bucket.
-func Open(cfg Config) (restic.Backend, error) {
-	return open(cfg)
+func Open(cfg Config, rt http.RoundTripper) (restic.Backend, error) {
+	return open(cfg, rt)
 }
 
 // Create opens the gs backend at the specified bucket and attempts to creates
@@ -104,54 +126,26 @@ func Open(cfg Config) (restic.Backend, error) {
 //
 // The service account must have the "storage.buckets.create" permission to
 // create a bucket the does not yet exist.
-func Create(cfg Config) (restic.Backend, error) {
-	be, err := open(cfg)
+func Create(cfg Config, rt http.RoundTripper) (restic.Backend, error) {
+	be, err := open(cfg, rt)
 	if err != nil {
 		return nil, errors.Wrap(err, "open")
 	}
 
 	// Try to determine if the bucket exists. If it does not, try to create it.
-	//
-	// A Get call has three typical error cases:
-	//
-	// * nil: Bucket exists and we have access to the metadata (returned).
-	//
-	// * 403: Bucket exists and we do not have access to the metadata. We
-	// don't have storage.buckets.get permission to the bucket, but we may
-	// still be able to access objects in the bucket.
-	//
-	// * 404: Bucket doesn't exist.
-	//
-	// Determining if the bucket is accessible is best-effort because the
-	// 403 case is ambiguous.
-	if _, err := be.service.Buckets.Get(be.bucketName).Do(); err != nil {
-		gerr, ok := err.(*googleapi.Error)
-		if !ok {
-			// Don't know what to do with this error.
-			return nil, errors.Wrap(err, "service.Buckets.Get")
+	ctx := context.Background()
+	exists, err := be.bucketExists(ctx, be.bucket)
+	if err != nil {
+		return nil, errors.Wrap(err, "service.Buckets.Get")
+	}
+
+	if !exists {
+		// Bucket doesn't exist, try to create it.
+		if err := be.bucket.Create(ctx, be.projectID, nil); err != nil {
+			// Always an error, as the bucket definitely doesn't exist.
+			return nil, errors.Wrap(err, "service.Buckets.Insert")
 		}
 
-		switch gerr.Code {
-		case 403:
-			// Bucket exists, but we don't know if it is
-			// accessible. Optimistically assume it is; if not,
-			// future Backend calls will fail.
-			debug.Log("Unable to determine if bucket %s is accessible (err %v). Continuing as if it is.", be.bucketName, err)
-		case 404:
-			// Bucket doesn't exist, try to create it.
-			bucket := &storage.Bucket{
-				Name: be.bucketName,
-			}
-
-			if _, err := be.service.Buckets.Insert(be.projectID, bucket).Do(); err != nil {
-				// Always an error, as the bucket definitely
-				// doesn't exist.
-				return nil, errors.Wrap(err, "service.Buckets.Insert")
-			}
-		default:
-			// Don't know what to do with this error.
-			return nil, errors.Wrap(err, "service.Buckets.Get")
-		}
 	}
 
 	return be, nil
@@ -195,7 +189,7 @@ func (be *Backend) Path() string {
 }
 
 // Save stores data in the backend at the handle.
-func (be *Backend) Save(ctx context.Context, h restic.Handle, rd io.Reader) (err error) {
+func (be *Backend) Save(ctx context.Context, h restic.Handle, rd restic.RewindReader) error {
 	if err := h.Valid(); err != nil {
 		return err
 	}
@@ -206,13 +200,6 @@ func (be *Backend) Save(ctx context.Context, h restic.Handle, rd io.Reader) (err
 
 	be.sem.GetToken()
 
-	// Check key does not already exist
-	if _, err := be.service.Objects.Get(be.bucketName, objName).Do(); err == nil {
-		debug.Log("%v already exists", h)
-		be.sem.ReleaseToken()
-		return errors.New("key already exists")
-	}
-
 	debug.Log("InsertObject(%v, %v)", be.bucketName, objName)
 
 	// Set chunk size to zero to disable resumable uploads.
@@ -240,12 +227,10 @@ func (be *Backend) Save(ctx context.Context, h restic.Handle, rd io.Reader) (err
 	//
 	// restic typically writes small blobs (4MB-30MB), so the resumable
 	// uploads are not providing significant benefit anyways.
-	cs := googleapi.ChunkSize(0)
-
-	info, err := be.service.Objects.Insert(be.bucketName,
-		&storage.Object{
-			Name: objName,
-		}).Media(rd, cs).Do()
+	w := be.bucket.Object(objName).NewWriter(ctx)
+	w.ChunkSize = 0
+	wbytes, err := io.Copy(w, rd)
+	w.Close()
 
 	be.sem.ReleaseToken()
 
@@ -254,7 +239,7 @@ func (be *Backend) Save(ctx context.Context, h restic.Handle, rd io.Reader) (err
 		return errors.Wrap(err, "service.Objects.Insert")
 	}
 
-	debug.Log("%v -> %v bytes", objName, info.Size)
+	debug.Log("%v -> %v bytes", objName, wbytes)
 	return nil
 }
 
@@ -270,10 +255,13 @@ func (wr wrapReader) Close() error {
 	return err
 }
 
-// Load returns a reader that yields the contents of the file at h at the
-// given offset. If length is nonzero, only a portion of the file is
-// returned. rd must be closed after use.
-func (be *Backend) Load(ctx context.Context, h restic.Handle, length int, offset int64) (io.ReadCloser, error) {
+// Load runs fn with a reader that yields the contents of the file at h at the
+// given offset.
+func (be *Backend) Load(ctx context.Context, h restic.Handle, length int, offset int64, fn func(rd io.Reader) error) error {
+	return backend.DefaultLoad(ctx, h, length, offset, be.openReader, fn)
+}
+
+func (be *Backend) openReader(ctx context.Context, h restic.Handle, length int, offset int64) (io.ReadCloser, error) {
 	debug.Log("Load %v, length %v, offset %v from %v", h, length, offset, be.Filename(h))
 	if err := h.Valid(); err != nil {
 		return nil, err
@@ -286,29 +274,23 @@ func (be *Backend) Load(ctx context.Context, h restic.Handle, length int, offset
 	if length < 0 {
 		return nil, errors.Errorf("invalid length %d", length)
 	}
+	if length == 0 {
+		// negative length indicates read till end to GCS lib
+		length = -1
+	}
 
 	objName := be.Filename(h)
 
 	be.sem.GetToken()
 
-	var byteRange string
-	if length > 0 {
-		byteRange = fmt.Sprintf("bytes=%d-%d", offset, offset+int64(length-1))
-	} else {
-		byteRange = fmt.Sprintf("bytes=%d-", offset)
-	}
-
-	req := be.service.Objects.Get(be.bucketName, objName)
-	// https://cloud.google.com/storage/docs/json_api/v1/parameters#range
-	req.Header().Set("Range", byteRange)
-	res, err := req.Download()
+	r, err := be.bucket.Object(objName).NewRangeReader(ctx, offset, int64(length))
 	if err != nil {
 		be.sem.ReleaseToken()
 		return nil, err
 	}
 
 	closeRd := wrapReader{
-		ReadCloser: res.Body,
+		ReadCloser: r,
 		f: func() {
 			debug.Log("Close()")
 			be.sem.ReleaseToken()
@@ -325,15 +307,15 @@ func (be *Backend) Stat(ctx context.Context, h restic.Handle) (bi restic.FileInf
 	objName := be.Filename(h)
 
 	be.sem.GetToken()
-	obj, err := be.service.Objects.Get(be.bucketName, objName).Do()
+	attr, err := be.bucket.Object(objName).Attrs(ctx)
 	be.sem.ReleaseToken()
 
 	if err != nil {
-		debug.Log("GetObject() err %v", err)
+		debug.Log("GetObjectAttributes() err %v", err)
 		return restic.FileInfo{}, errors.Wrap(err, "service.Objects.Get")
 	}
 
-	return restic.FileInfo{Size: int64(obj.Size)}, nil
+	return restic.FileInfo{Size: attr.Size, Name: h.Name}, nil
 }
 
 // Test returns true if a blob of the given type and name exists in the backend.
@@ -342,7 +324,7 @@ func (be *Backend) Test(ctx context.Context, h restic.Handle) (bool, error) {
 	objName := be.Filename(h)
 
 	be.sem.GetToken()
-	_, err := be.service.Objects.Get(be.bucketName, objName).Do()
+	_, err := be.bucket.Object(objName).Attrs(ctx)
 	be.sem.ReleaseToken()
 
 	if err == nil {
@@ -357,88 +339,78 @@ func (be *Backend) Remove(ctx context.Context, h restic.Handle) error {
 	objName := be.Filename(h)
 
 	be.sem.GetToken()
-	err := be.service.Objects.Delete(be.bucketName, objName).Do()
+	err := be.bucket.Object(objName).Delete(ctx)
 	be.sem.ReleaseToken()
 
-	if er, ok := err.(*googleapi.Error); ok {
-		if er.Code == 404 {
-			err = nil
-		}
+	if err == storage.ErrObjectNotExist {
+		err = nil
 	}
 
 	debug.Log("Remove(%v) at %v -> err %v", h, objName, err)
 	return errors.Wrap(err, "client.RemoveObject")
 }
 
-// List returns a channel that yields all names of blobs of type t. A
-// goroutine is started for this. If the channel done is closed, sending
-// stops.
-func (be *Backend) List(ctx context.Context, t restic.FileType) <-chan string {
+// List runs fn for each file in the backend which has the type t. When an
+// error occurs (or fn returns an error), List stops and returns it.
+func (be *Backend) List(ctx context.Context, t restic.FileType, fn func(restic.FileInfo) error) error {
 	debug.Log("listing %v", t)
-	ch := make(chan string)
 
 	prefix, _ := be.Basedir(t)
 
 	// make sure prefix ends with a slash
-	if prefix[len(prefix)-1] != '/' {
+	if !strings.HasSuffix(prefix, "/") {
 		prefix += "/"
 	}
 
-	go func() {
-		defer close(ch)
+	ctx, cancel := context.WithCancel(ctx)
+	defer cancel()
 
-		listReq := be.service.Objects.List(be.bucketName).Prefix(prefix).MaxResults(int64(be.listMaxItems))
-		for {
-			be.sem.GetToken()
-			obj, err := listReq.Do()
-			be.sem.ReleaseToken()
+	itr := be.bucket.Objects(ctx, &storage.Query{Prefix: prefix})
 
-			if err != nil {
-				fmt.Fprintf(os.Stderr, "error listing %v: %v\n", prefix, err)
-				return
-			}
-
-			debug.Log("returned %v items", len(obj.Items))
-
-			for _, item := range obj.Items {
-				m := strings.TrimPrefix(item.Name, prefix)
-				if m == "" {
-					continue
-				}
-
-				select {
-				case ch <- path.Base(m):
-				case <-ctx.Done():
-					return
-				}
-			}
-
-			if obj.NextPageToken == "" {
-				break
-			}
-			listReq.PageToken(obj.NextPageToken)
+	for {
+		be.sem.GetToken()
+		attrs, err := itr.Next()
+		be.sem.ReleaseToken()
+		if err == iterator.Done {
+			break
+		}
+		if err != nil {
+			return err
+		}
+		m := strings.TrimPrefix(attrs.Name, prefix)
+		if m == "" {
+			continue
 		}
-	}()
 
-	return ch
+		fi := restic.FileInfo{
+			Name: path.Base(m),
+			Size: int64(attrs.Size),
+		}
+
+		err = fn(fi)
+		if err != nil {
+			return err
+		}
+
+		if ctx.Err() != nil {
+			return ctx.Err()
+		}
+	}
+
+	return ctx.Err()
 }
 
 // Remove keys for a specified backend type.
 func (be *Backend) removeKeys(ctx context.Context, t restic.FileType) error {
-	for key := range be.List(ctx, restic.DataFile) {
-		err := be.Remove(ctx, restic.Handle{Type: restic.DataFile, Name: key})
-		if err != nil {
-			return err
-		}
-	}
-
-	return nil
+	return be.List(ctx, t, func(fi restic.FileInfo) error {
+		return be.Remove(ctx, restic.Handle{Type: t, Name: fi.Name})
+	})
 }
 
 // Delete removes all restic keys in the bucket. It will not remove the bucket itself.
 func (be *Backend) Delete(ctx context.Context) error {
 	alltypes := []restic.FileType{
-		restic.DataFile,
+		restic.PackFile,
 		restic.KeyFile,
 		restic.LockFile,
 		restic.SnapshotFile,
