@@ -5,6 +5,7 @@ import (
 	"io"
 	"os"
 	"path/filepath"
+	"syscall"
 
 	"github.com/restic/restic/internal/errors"
 	"github.com/restic/restic/internal/restic"
@@ -25,62 +26,23 @@ var _ restic.Backend = &Local{}
 
 const defaultLayout = "default"
 
-// dirExists returns true if the name exists and is a directory.
-func dirExists(name string) bool {
-	f, err := fs.Open(name)
-	if err != nil {
-		return false
-	}
-
-	fi, err := f.Stat()
-	if err != nil {
-		return false
-	}
-
-	if err = f.Close(); err != nil {
-		return false
-	}
-
-	return fi.IsDir()
-}
-
 // Open opens the local backend as specified by config.
-func Open(cfg Config) (*Local, error) {
+func Open(ctx context.Context, cfg Config) (*Local, error) {
 	debug.Log("open local backend at %v (layout %q)", cfg.Path, cfg.Layout)
-	l, err := backend.ParseLayout(&backend.LocalFilesystem{}, cfg.Layout, defaultLayout, cfg.Path)
+	l, err := backend.ParseLayout(ctx, &backend.LocalFilesystem{}, cfg.Layout, defaultLayout, cfg.Path)
 	if err != nil {
 		return nil, err
 	}
 
-	be := &Local{Config: cfg, Layout: l}
-
-	// if data dir exists, make sure that all subdirs also exist
-	datadir := be.Dirname(restic.Handle{Type: restic.DataFile})
-	if dirExists(datadir) {
-		debug.Log("datadir %v exists", datadir)
-		for _, d := range be.Paths() {
-			if !fs.HasPathPrefix(datadir, d) {
-				debug.Log("%v is not subdir of datadir %v", d, datadir)
-				continue
-			}
-
-			debug.Log("MkdirAll %v", d)
-			err := fs.MkdirAll(d, backend.Modes.Dir)
-			if err != nil {
-				return nil, errors.Wrap(err, "MkdirAll")
-			}
-		}
-	}
-
-	return be, nil
+	return &Local{Config: cfg, Layout: l}, nil
 }
 
 // Create creates all the necessary files and directories for a new local
 // backend at dir. Afterwards a new config blob should be created.
-func Create(cfg Config) (*Local, error) {
+func Create(ctx context.Context, cfg Config) (*Local, error) {
 	debug.Log("create local backend at %v (layout %q)", cfg.Path, cfg.Layout)
 
-	l, err := backend.ParseLayout(&backend.LocalFilesystem{}, cfg.Layout, defaultLayout, cfg.Path)
+	l, err := backend.ParseLayout(ctx, &backend.LocalFilesystem{}, cfg.Layout, defaultLayout, cfg.Path)
 	if err != nil {
 		return nil, err
 	}
@@ -118,26 +80,30 @@ func (b *Local) IsNotExist(err error) bool {
 }
 
 // Save stores data in the backend at the handle.
-func (b *Local) Save(ctx context.Context, h restic.Handle, rd io.Reader) error {
+func (b *Local) Save(ctx context.Context, h restic.Handle, rd restic.RewindReader) error {
 	debug.Log("Save %v", h)
 	if err := h.Valid(); err != nil {
 		return err
 	}
 
-	if h.Type == restic.LockFile {
-		lockDir := b.Dirname(h)
-		if !dirExists(lockDir) {
-			debug.Log("locks/ does not exist yet, creating now.")
-			if err := fs.MkdirAll(lockDir, backend.Modes.Dir); err != nil {
-				return errors.Wrap(err, "MkdirAll")
-			}
-		}
-	}
-
 	filename := b.Filename(h)
 
 	// create new file
 	f, err := fs.OpenFile(filename, os.O_CREATE|os.O_EXCL|os.O_WRONLY, backend.Modes.File)
+
+	if b.IsNotExist(err) {
+		debug.Log("error %v: creating dir", err)
+
+		// error is caused by a missing directory, try to create it
+		mkdirErr := os.MkdirAll(filepath.Dir(filename), backend.Modes.Dir)
+		if mkdirErr != nil {
+			debug.Log("error creating dir %v: %v", filepath.Dir(filename), mkdirErr)
+		} else {
+			// try again
+			f, err = fs.OpenFile(filename, os.O_CREATE|os.O_EXCL|os.O_WRONLY, backend.Modes.File)
+		}
+	}
+
 	if err != nil {
 		return errors.Wrap(err, "OpenFile")
 	}
@@ -150,8 +116,13 @@ func (b *Local) Save(ctx context.Context, h restic.Handle, rd io.Reader) error {
 	}
 
 	if err = f.Sync(); err != nil {
-		_ = f.Close()
-		return errors.Wrap(err, "Sync")
+		pathErr, ok := err.(*os.PathError)
+		isNotSupported := ok && pathErr.Op == "sync" && pathErr.Err == syscall.ENOTSUP
+		// ignore error if filesystem does not support the sync operation
+		if !isNotSupported {
+			_ = f.Close()
+			return errors.Wrap(err, "Sync")
+		}
 	}
 
 	err = f.Close()
@@ -159,13 +130,24 @@ func (b *Local) Save(ctx context.Context, h restic.Handle, rd io.Reader) error {
 		return errors.Wrap(err, "Close")
 	}
 
-	return setNewFileMode(filename, backend.Modes.File)
+	// try to mark file as read-only to avoid accidential modifications
+	// ignore if the operation fails as some filesystems don't allow the chmod call
+	// e.g. exfat and network file systems with certain mount options
+	err = setFileReadonly(filename, backend.Modes.File)
+	if err != nil && !os.IsPermission(err) {
+		return errors.Wrap(err, "Chmod")
+	}
+
+	return nil
 }
 
-// Load returns a reader that yields the contents of the file at h at the
-// given offset. If length is nonzero, only a portion of the file is
-// returned. rd must be closed after use.
-func (b *Local) Load(ctx context.Context, h restic.Handle, length int, offset int64) (io.ReadCloser, error) {
+// Load runs fn with a reader that yields the contents of the file at h at the
+// given offset.
+func (b *Local) Load(ctx context.Context, h restic.Handle, length int, offset int64, fn func(rd io.Reader) error) error {
+	return backend.DefaultLoad(ctx, h, length, offset, b.openReader, fn)
+}
+
+func (b *Local) openReader(ctx context.Context, h restic.Handle, length int, offset int64) (io.ReadCloser, error) {
 	debug.Log("Load %v, length %v, offset %v", h, length, offset)
 	if err := h.Valid(); err != nil {
 		return nil, err
@@ -207,7 +189,7 @@ func (b *Local) Stat(ctx context.Context, h restic.Handle) (restic.FileInfo, err
 		return restic.FileInfo{}, errors.Wrap(err, "Stat")
 	}
 
-	return restic.FileInfo{Size: fi.Size()}, nil
+	return restic.FileInfo{Size: fi.Size(), Name: h.Name}, nil
 }
 
 // Test returns true if a blob of the given type and name exists in the backend.
@@ -231,7 +213,7 @@ func (b *Local) Remove(ctx context.Context, h restic.Handle) error {
 
 	// reset read-only flag
 	err := fs.Chmod(fn, 0666)
-	if err != nil {
+	if err != nil && !os.IsPermission(err) {
 		return errors.Wrap(err, "Chmod")
 	}
 
@@ -242,52 +224,55 @@ func isFile(fi os.FileInfo) bool {
 	return fi.Mode()&(os.ModeType|os.ModeCharDevice) == 0
 }
 
-// List returns a channel that yields all names of blobs of type t. A
-// goroutine is started for this.
-func (b *Local) List(ctx context.Context, t restic.FileType) <-chan string {
+// List runs fn for each file in the backend which has the type t. When an
+// error occurs (or fn returns an error), List stops and returns it.
+func (b *Local) List(ctx context.Context, t restic.FileType, fn func(restic.FileInfo) error) error {
 	debug.Log("List %v", t)
 
-	ch := make(chan string)
-
-	go func() {
-		defer close(ch)
-
-		basedir, subdirs := b.Basedir(t)
-		err := fs.Walk(basedir, func(path string, fi os.FileInfo, err error) error {
-			debug.Log("walk on %v, %v\n", path, fi.IsDir())
-			if err != nil {
-				return err
-			}
-
-			if path == basedir {
-				return nil
-			}
-
-			if !isFile(fi) {
-				return nil
-			}
-
-			if fi.IsDir() && !subdirs {
-				return filepath.SkipDir
-			}
-
-			debug.Log("send %v\n", filepath.Base(path))
-
-			select {
-			case ch <- filepath.Base(path):
-			case <-ctx.Done():
-				return nil
-			}
-
-			return nil
-		})
-
+	basedir, subdirs := b.Basedir(t)
+	err := fs.Walk(basedir, func(path string, fi os.FileInfo, err error) error {
+		debug.Log("walk on %v\n", path)
 		if err != nil {
-			debug.Log("Walk %v", err)
+			return err
 		}
-	}()
 
-	return ch
+		if path == basedir {
+			return nil
+		}
+
+		if !isFile(fi) {
+			return nil
+		}
+
+		if fi.IsDir() && !subdirs {
+			return filepath.SkipDir
+		}
+
+		debug.Log("send %v\n", filepath.Base(path))
+
+		rfi := restic.FileInfo{
+			Name: filepath.Base(path),
+			Size: fi.Size(),
+		}
+
+		if ctx.Err() != nil {
+			return ctx.Err()
+		}
+
+		err = fn(rfi)
+		if err != nil {
+			return err
+		}
+
+		return ctx.Err()
+	})
+
+	if b.IsNotExist(err) {
+		debug.Log("ignoring non-existing directory")
+		return nil
+	}
+
+	return err
 }
 
 // Delete removes the repository and all files.
