@@ -8,7 +8,6 @@ import (
 	"os"
 	"os/exec"
 	"path"
-	"strings"
 	"time"
 
 	"github.com/restic/restic/internal/errors"
@@ -36,7 +35,7 @@ var _ restic.Backend = &SFTP{}
 
 const defaultLayout = "default"
 
-func startClient(preExec, postExec func(), program string, args ...string) (*SFTP, error) {
+func startClient(program string, args ...string) (*SFTP, error) {
 	debug.Log("start client %v %v", program, args)
 	// Connect to a remote host and request the sftp subsystem via the 'ssh'
 	// command.  This assumes that passwordless login is correctly configured.
@@ -65,25 +64,19 @@ func startClient(preExec, postExec func(), program string, args ...string) (*SFT
 		return nil, errors.Wrap(err, "cmd.StdoutPipe")
 	}
 
-	if preExec != nil {
-		preExec()
-	}
-
-	// start the process
-	if err := cmd.Start(); err != nil {
+	bg, err := backend.StartForeground(cmd)
+	if err != nil {
 		return nil, errors.Wrap(err, "cmd.Start")
 	}
 
-	if postExec != nil {
-		postExec()
-	}
-
 	// wait in a different goroutine
 	ch := make(chan error, 1)
 	go func() {
 		err := cmd.Wait()
 		debug.Log("ssh command exited, err %v", err)
-		ch <- errors.Wrap(err, "cmd.Wait")
+		for {
+			ch <- errors.Wrap(err, "ssh command exited")
+		}
 	}()
 
 	// open the SFTP session
@@ -92,6 +85,11 @@ func startClient(preExec, postExec func(), program string, args ...string) (*SFT
 		return nil, errors.Errorf("unable to start the sftp session, error: %v", err)
 	}
 
+	err = bg()
+	if err != nil {
+		return nil, errors.Wrap(err, "bg")
+	}
+
 	return &SFTP{c: client, cmd: cmd, result: ch}, nil
 }
 
@@ -111,7 +109,7 @@ func (r *SFTP) clientError() error {
 // Open opens an sftp backend as described by the config by running
 // "ssh" with the appropriate arguments (or cfg.Command, if set). The function
 // preExec is run just before, postExec just after starting a program.
-func Open(cfg Config, preExec, postExec func()) (*SFTP, error) {
+func Open(ctx context.Context, cfg Config) (*SFTP, error) {
 	debug.Log("open backend with config %#v", cfg)
 
 	cmd, args, err := buildSSHCommand(cfg)
@@ -119,65 +117,27 @@ func Open(cfg Config, preExec, postExec func()) (*SFTP, error) {
 		return nil, err
 	}
 
-	sftp, err := startClient(preExec, postExec, cmd, args...)
+	sftp, err := startClient(cmd, args...)
 	if err != nil {
 		debug.Log("unable to start program: %v", err)
 		return nil, err
 	}
 
-	sftp.Layout, err = backend.ParseLayout(sftp, cfg.Layout, defaultLayout, cfg.Path)
+	sftp.Layout, err = backend.ParseLayout(ctx, sftp, cfg.Layout, defaultLayout, cfg.Path)
 	if err != nil {
 		return nil, err
 	}
 
 	debug.Log("layout: %v\n", sftp.Layout)
 
-	if err := sftp.checkDataSubdirs(); err != nil {
-		debug.Log("checkDataSubdirs returned %v", err)
-		return nil, err
-	}
-
 	sftp.Config = cfg
 	sftp.p = cfg.Path
 	return sftp, nil
 }
 
-func (r *SFTP) checkDataSubdirs() error {
-	datadir := r.Dirname(restic.Handle{Type: restic.DataFile})
-
-	// check if all paths for data/ exist
-	entries, err := r.ReadDir(datadir)
-	if r.IsNotExist(err) {
-		return nil
-	}
-
-	if err != nil {
-		return err
-	}
-
-	subdirs := make(map[string]struct{}, len(entries))
-	for _, entry := range entries {
-		subdirs[entry.Name()] = struct{}{}
-	}
-
-	for i := 0; i < 256; i++ {
-		subdir := fmt.Sprintf("%02x", i)
-		if _, ok := subdirs[subdir]; !ok {
-			debug.Log("subdir %v is missing, creating", subdir)
-			err := r.mkdirAll(path.Join(datadir, subdir), backend.Modes.Dir)
-			if err != nil {
-				return err
-			}
-		}
-	}
-
-	return nil
-}
-
 func (r *SFTP) mkdirAllDataSubdirs() error {
 	for _, d := range r.Paths() {
-		err := r.mkdirAll(d, backend.Modes.Dir)
-		debug.Log("mkdirAll %v -> %v", d, err)
+		err := r.c.MkdirAll(d)
 		if err != nil {
 			return err
 		}
@@ -192,7 +152,7 @@ func (r *SFTP) Join(p ...string) string {
 }
 
 // ReadDir returns the entries for a directory.
-func (r *SFTP) ReadDir(dir string) ([]os.FileInfo, error) {
+func (r *SFTP) ReadDir(ctx context.Context, dir string) ([]os.FileInfo, error) {
 	fi, err := r.c.ReadDir(dir)
 
 	// sftp client does not specify dir name on error, so add it here
@@ -203,6 +163,8 @@ func (r *SFTP) ReadDir(dir string) ([]os.FileInfo, error) {
 
 // IsNotExist returns true if the error is caused by a not existing file.
 func (r *SFTP) IsNotExist(err error) bool {
+	err = errors.Cause(err)
+
 	if os.IsNotExist(err) {
 		return true
 	}
@@ -217,15 +179,21 @@ func (r *SFTP) IsNotExist(err error) bool {
 
 func buildSSHCommand(cfg Config) (cmd string, args []string, err error) {
 	if cfg.Command != "" {
-		return SplitShellArgs(cfg.Command)
+		args, err := backend.SplitShellStrings(cfg.Command)
+		if err != nil {
+			return "", nil, err
+		}
+
+		return args[0], args[1:], nil
 	}
 
 	cmd = "ssh"
 
-	hostport := strings.Split(cfg.Host, ":")
-	args = []string{hostport[0]}
-	if len(hostport) > 1 {
-		args = append(args, "-p", hostport[1])
+	host, port := cfg.Host, cfg.Port
+
+	args = []string{host}
+	if port != "" {
+		args = append(args, "-p", port)
 	}
 	if cfg.User != "" {
 		args = append(args, "-l")
@@ -239,19 +207,19 @@ func buildSSHCommand(cfg Config) (cmd string, args []string, err error) {
 // Create creates an sftp backend as described by the config by running "ssh"
 // with the appropriate arguments (or cfg.Command, if set). The function
 // preExec is run just before, postExec just after starting a program.
-func Create(cfg Config, preExec, postExec func()) (*SFTP, error) {
+func Create(ctx context.Context, cfg Config) (*SFTP, error) {
 	cmd, args, err := buildSSHCommand(cfg)
 	if err != nil {
 		return nil, err
 	}
 
-	sftp, err := startClient(preExec, postExec, cmd, args...)
+	sftp, err := startClient(cmd, args...)
 	if err != nil {
 		debug.Log("unable to start program: %v", err)
 		return nil, err
 	}
 
-	sftp.Layout, err = backend.ParseLayout(sftp, cfg.Layout, defaultLayout, cfg.Path)
+	sftp.Layout, err = backend.ParseLayout(ctx, sftp, cfg.Layout, defaultLayout, cfg.Path)
 	if err != nil {
 		return nil, err
 	}
@@ -273,7 +241,7 @@ func Create(cfg Config, preExec, postExec func()) (*SFTP, error) {
 	}
 
 	// open backend
-	return Open(cfg, preExec, postExec)
+	return Open(ctx, cfg)
 }
 
 // Location returns this backend's location (the directory name).
@@ -281,38 +249,6 @@ func (r *SFTP) Location() string {
 	return r.p
 }
 
-func (r *SFTP) mkdirAll(dir string, mode os.FileMode) error {
-	// check if directory already exists
-	fi, err := r.c.Lstat(dir)
-	if err == nil {
-		if fi.IsDir() {
-			return nil
-		}
-
-		return errors.Errorf("mkdirAll(%s): entry exists but is not a directory", dir)
-	}
-
-	// create parent directories
-	errMkdirAll := r.mkdirAll(path.Dir(dir), backend.Modes.Dir)
-
-	// create directory
-	errMkdir := r.c.Mkdir(dir)
-
-	// test if directory was created successfully
-	fi, err = r.c.Lstat(dir)
-	if err != nil {
-		// return previous errors
-		return errors.Errorf("mkdirAll(%s): unable to create directories: %v, %v", dir, errMkdirAll, errMkdir)
-	}
-
-	if !fi.IsDir() {
-		return errors.Errorf("mkdirAll(%s): entry exists but is not a directory", dir)
-	}
-
-	// set mode
-	return r.c.Chmod(dir, mode)
-}
-
 // Join joins the given paths and cleans them afterwards. This always uses
 // forward slashes, which is required by sftp.
 func Join(parts ...string) string {
@@ -320,7 +256,7 @@ func Join(parts ...string) string {
 }
 
 // Save stores data in the backend at the handle.
-func (r *SFTP) Save(ctx context.Context, h restic.Handle, rd io.Reader) (err error) {
+func (r *SFTP) Save(ctx context.Context, h restic.Handle, rd restic.RewindReader) error {
 	debug.Log("Save %v", h)
 	if err := r.clientError(); err != nil {
 		return err
@@ -334,14 +270,16 @@ func (r *SFTP) Save(ctx context.Context, h restic.Handle, rd io.Reader) (err err
 
 	// create new file
 	f, err := r.c.OpenFile(filename, os.O_CREATE|os.O_EXCL|os.O_WRONLY)
-	if r.IsNotExist(errors.Cause(err)) {
-		// create the locks dir, then try again
-		err = r.mkdirAll(r.Dirname(h), backend.Modes.Dir)
-		if err != nil {
-			return errors.Wrap(err, "MkdirAll")
-		}
 
-		return r.Save(ctx, h, rd)
+	if r.IsNotExist(err) {
+		// error is caused by a missing directory, try to create it
+		mkdirErr := r.c.MkdirAll(r.Dirname(h))
+		if mkdirErr != nil {
+			debug.Log("error creating dir %v: %v", r.Dirname(h), mkdirErr)
+		} else {
+			// try again
+			f, err = r.c.OpenFile(filename, os.O_CREATE|os.O_EXCL|os.O_WRONLY)
+		}
 	}
 
 	if err != nil {
@@ -363,10 +301,13 @@ func (r *SFTP) Save(ctx context.Context, h restic.Handle, rd io.Reader) (err err
 	return errors.Wrap(r.c.Chmod(filename, backend.Modes.File), "Chmod")
 }
 
-// Load returns a reader that yields the contents of the file at h at the
-// given offset. If length is nonzero, only a portion of the file is
-// returned. rd must be closed after use.
-func (r *SFTP) Load(ctx context.Context, h restic.Handle, length int, offset int64) (io.ReadCloser, error) {
+// Load runs fn with a reader that yields the contents of the file at h at the
+// given offset.
+func (r *SFTP) Load(ctx context.Context, h restic.Handle, length int, offset int64, fn func(rd io.Reader) error) error {
+	return backend.DefaultLoad(ctx, h, length, offset, r.openReader, fn)
+}
+
+func (r *SFTP) openReader(ctx context.Context, h restic.Handle, length int, offset int64) (io.ReadCloser, error) {
 	debug.Log("Load %v, length %v, offset %v", h, length, offset)
 	if err := h.Valid(); err != nil {
 		return nil, err
@@ -412,7 +353,7 @@ func (r *SFTP) Stat(ctx context.Context, h restic.Handle) (restic.FileInfo, erro
 		return restic.FileInfo{}, errors.Wrap(err, "Lstat")
 	}
 
-	return restic.FileInfo{Size: fi.Size()}, nil
+	return restic.FileInfo{Size: fi.Size(), Name: h.Name}, nil
 }
 
 // Test returns true if a blob of the given type and name exists in the backend.
@@ -444,47 +385,58 @@ func (r *SFTP) Remove(ctx context.Context, h restic.Handle) error {
 	return r.c.Remove(r.Filename(h))
 }
 
-// List returns a channel that yields all names of blobs of type t. A
-// goroutine is started for this. If the channel done is closed, sending
-// stops.
-func (r *SFTP) List(ctx context.Context, t restic.FileType) <-chan string {
+// List runs fn for each file in the backend which has the type t. When an
+// error occurs (or fn returns an error), List stops and returns it.
+func (r *SFTP) List(ctx context.Context, t restic.FileType, fn func(restic.FileInfo) error) error {
 	debug.Log("List %v", t)
 
-	ch := make(chan string)
-
-	go func() {
-		defer close(ch)
-
-		basedir, subdirs := r.Basedir(t)
-		walker := r.c.Walk(basedir)
-		for walker.Step() {
-			if walker.Err() != nil {
-				continue
-			}
-
-			if walker.Path() == basedir {
-				continue
-			}
-
-			if walker.Stat().IsDir() && !subdirs {
-				walker.SkipDir()
-				continue
-			}
-
-			if !walker.Stat().Mode().IsRegular() {
-				continue
-			}
-
-			select {
-			case ch <- path.Base(walker.Path()):
-			case <-ctx.Done():
-				return
+	basedir, subdirs := r.Basedir(t)
+	walker := r.c.Walk(basedir)
+	for walker.Step() {
+		if walker.Err() != nil {
+			if r.IsNotExist(walker.Err()) {
+				debug.Log("ignoring non-existing directory")
+				return nil
 			}
+			return walker.Err()
 		}
-	}()
 
-	return ch
+		if walker.Path() == basedir {
+			continue
+		}
 
+		if walker.Stat().IsDir() && !subdirs {
+			walker.SkipDir()
+			continue
+		}
+
+		fi := walker.Stat()
+		if !fi.Mode().IsRegular() {
+			continue
+		}
+
+		debug.Log("send %v\n", path.Base(walker.Path()))
+
+		rfi := restic.FileInfo{
+			Name: path.Base(walker.Path()),
+			Size: fi.Size(),
+		}
+
+		if ctx.Err() != nil {
+			return ctx.Err()
+		}
+
+		err := fn(rfi)
+		if err != nil {
+			return err
+		}
+
+		if ctx.Err() != nil {
+			return ctx.Err()
+		}
+	}
+
+	return ctx.Err()
 }
 
 var closeTimeout = 2 * time.Second
@@ -515,8 +467,8 @@ func (r *SFTP) Close() error {
 	return nil
 }
 
-func (r *SFTP) deleteRecursive(name string) error {
-	entries, err := r.ReadDir(name)
+func (r *SFTP) deleteRecursive(ctx context.Context, name string) error {
+	entries, err := r.ReadDir(ctx, name)
 	if err != nil {
 		return errors.Wrap(err, "ReadDir")
 	}
@@ -524,7 +476,7 @@ func (r *SFTP) deleteRecursive(name string) error {
 	for _, fi := range entries {
 		itemName := r.Join(name, fi.Name())
 		if fi.IsDir() {
-			err := r.deleteRecursive(itemName)
+			err := r.deleteRecursive(ctx, itemName)
 			if err != nil {
 				return errors.Wrap(err, "ReadDir")
 			}
@@ -547,6 +499,6 @@ func (r *SFTP) deleteRecursive(name string) error {
 }
 
 // Delete removes all data in the backend.
-func (r *SFTP) Delete(context.Context) error {
-	return r.deleteRecursive(r.p)
+func (r *SFTP) Delete(ctx context.Context) error {
+	return r.deleteRecursive(ctx, r.p)
 }
