@@ -6,7 +6,6 @@ import (
 	"io"
 	"net/http"
 	"path"
-	"path/filepath"
 	"strings"
 	"time"
 
@@ -18,8 +17,6 @@ import (
 	"github.com/ncw/swift"
 )
 
-const connLimit = 10
-
 // beSwift is a backend which stores the data on a swift endpoint.
 type beSwift struct {
 	conn      *swift.Connection
@@ -44,19 +41,22 @@ func Open(cfg Config, rt http.RoundTripper) (restic.Backend, error) {
 
 	be := &beSwift{
 		conn: &swift.Connection{
-			UserName:       cfg.UserName,
-			Domain:         cfg.Domain,
-			ApiKey:         cfg.APIKey,
-			AuthUrl:        cfg.AuthURL,
-			Region:         cfg.Region,
-			Tenant:         cfg.Tenant,
-			TenantId:       cfg.TenantID,
-			TenantDomain:   cfg.TenantDomain,
-			TrustId:        cfg.TrustID,
-			StorageUrl:     cfg.StorageURL,
-			AuthToken:      cfg.AuthToken,
-			ConnectTimeout: time.Minute,
-			Timeout:        time.Minute,
+			UserName:                    cfg.UserName,
+			Domain:                      cfg.Domain,
+			ApiKey:                      cfg.APIKey,
+			AuthUrl:                     cfg.AuthURL,
+			Region:                      cfg.Region,
+			Tenant:                      cfg.Tenant,
+			TenantId:                    cfg.TenantID,
+			TenantDomain:                cfg.TenantDomain,
+			TrustId:                     cfg.TrustID,
+			StorageUrl:                  cfg.StorageURL,
+			AuthToken:                   cfg.AuthToken,
+			ApplicationCredentialId:     cfg.ApplicationCredentialID,
+			ApplicationCredentialName:   cfg.ApplicationCredentialName,
+			ApplicationCredentialSecret: cfg.ApplicationCredentialSecret,
+			ConnectTimeout:              time.Minute,
+			Timeout:                     time.Minute,
 
 			Transport: rt,
 		},
@@ -110,10 +110,13 @@ func (be *beSwift) Location() string {
 	return be.container
 }
 
-// Load returns a reader that yields the contents of the file at h at the
-// given offset. If length is nonzero, only a portion of the file is
-// returned. rd must be closed after use.
-func (be *beSwift) Load(ctx context.Context, h restic.Handle, length int, offset int64) (io.ReadCloser, error) {
+// Load runs fn with a reader that yields the contents of the file at h at the
+// given offset.
+func (be *beSwift) Load(ctx context.Context, h restic.Handle, length int, offset int64, fn func(rd io.Reader) error) error {
+	return backend.DefaultLoad(ctx, h, length, offset, be.openReader, fn)
+}
+
+func (be *beSwift) openReader(ctx context.Context, h restic.Handle, length int, offset int64) (io.ReadCloser, error) {
 	debug.Log("Load %v, length %v, offset %v", h, length, offset)
 	if err := h.Valid(); err != nil {
 		return nil, err
@@ -154,8 +157,8 @@ func (be *beSwift) Load(ctx context.Context, h restic.Handle, length int, offset
 }
 
 // Save stores data in the backend at the handle.
-func (be *beSwift) Save(ctx context.Context, h restic.Handle, rd io.Reader) (err error) {
-	if err = h.Valid(); err != nil {
+func (be *beSwift) Save(ctx context.Context, h restic.Handle, rd restic.RewindReader) error {
+	if err := h.Valid(); err != nil {
 		return err
 	}
 
@@ -166,23 +169,10 @@ func (be *beSwift) Save(ctx context.Context, h restic.Handle, rd io.Reader) (err
 	be.sem.GetToken()
 	defer be.sem.ReleaseToken()
 
-	// Check key does not already exist
-	switch _, _, err = be.conn.Object(be.container, objName); err {
-	case nil:
-		debug.Log("%v already exists", h)
-		return errors.New("key already exists")
-
-	case swift.ObjectNotFound:
-		// Ok, that's what we want
-
-	default:
-		return errors.Wrap(err, "conn.Object")
-	}
-
 	encoding := "binary/octet-stream"
 
 	debug.Log("PutObject(%v, %v, %v)", be.container, objName, encoding)
-	_, err = be.conn.ObjectPut(be.container, objName, rd, true, "", encoding, nil)
+	_, err := be.conn.ObjectPut(be.container, objName, rd, true, "", encoding, nil)
 	debug.Log("%v, err %#v", objName, err)
 
 	return errors.Wrap(err, "client.PutObject")
@@ -203,7 +193,7 @@ func (be *beSwift) Stat(ctx context.Context, h restic.Handle) (bi restic.FileInf
 		return restic.FileInfo{}, errors.Wrap(err, "conn.Object")
 	}
 
-	return restic.FileInfo{Size: obj.Bytes}, nil
+	return restic.FileInfo{Size: obj.Bytes, Name: h.Name}, nil
 }
 
 // Test returns true if a blob of the given type and name exists in the backend.
@@ -237,61 +227,62 @@ func (be *beSwift) Remove(ctx context.Context, h restic.Handle) error {
 	return errors.Wrap(err, "conn.ObjectDelete")
 }
 
-// List returns a channel that yields all names of blobs of type t. A
-// goroutine is started for this. If the channel done is closed, sending
-// stops.
-func (be *beSwift) List(ctx context.Context, t restic.FileType) <-chan string {
+// List runs fn for each file in the backend which has the type t. When an
+// error occurs (or fn returns an error), List stops and returns it.
+func (be *beSwift) List(ctx context.Context, t restic.FileType, fn func(restic.FileInfo) error) error {
 	debug.Log("listing %v", t)
-	ch := make(chan string)
 
 	prefix, _ := be.Basedir(t)
 	prefix += "/"
 
-	go func() {
-		defer close(ch)
+	err := be.conn.ObjectsWalk(be.container, &swift.ObjectsOpts{Prefix: prefix},
+		func(opts *swift.ObjectsOpts) (interface{}, error) {
+			be.sem.GetToken()
+			newObjects, err := be.conn.Objects(be.container, opts)
+			be.sem.ReleaseToken()
 
-		err := be.conn.ObjectsWalk(be.container, &swift.ObjectsOpts{Prefix: prefix},
-			func(opts *swift.ObjectsOpts) (interface{}, error) {
-				be.sem.GetToken()
-				newObjects, err := be.conn.ObjectNames(be.container, opts)
-				be.sem.ReleaseToken()
+			if err != nil {
+				return nil, errors.Wrap(err, "conn.ObjectNames")
+			}
+			for _, obj := range newObjects {
+				m := path.Base(strings.TrimPrefix(obj.Name, prefix))
+				if m == "" {
+					continue
+				}
 
+				fi := restic.FileInfo{
+					Name: m,
+					Size: obj.Bytes,
+				}
+
+				if ctx.Err() != nil {
+					return nil, ctx.Err()
+				}
+
+				err := fn(fi)
 				if err != nil {
-					return nil, errors.Wrap(err, "conn.ObjectNames")
+					return nil, err
 				}
-				for _, obj := range newObjects {
-					m := filepath.Base(strings.TrimPrefix(obj, prefix))
-					if m == "" {
-						continue
-					}
 
-					select {
-					case ch <- m:
-					case <-ctx.Done():
-						return nil, io.EOF
-					}
+				if ctx.Err() != nil {
+					return nil, ctx.Err()
 				}
-				return newObjects, nil
-			})
+			}
+			return newObjects, nil
+		})
 
-		if err != nil {
-			debug.Log("ObjectsWalk returned error: %v", err)
-		}
-	}()
+	if err != nil {
+		return err
+	}
 
-	return ch
+	return ctx.Err()
 }
 
 // Remove keys for a specified backend type.
 func (be *beSwift) removeKeys(ctx context.Context, t restic.FileType) error {
-	for key := range be.List(ctx, t) {
-		err := be.Remove(ctx, restic.Handle{Type: t, Name: key})
-		if err != nil {
-			return err
-		}
-	}
-
-	return nil
+	return be.List(ctx, t, func(fi restic.FileInfo) error {
+		return be.Remove(ctx, restic.Handle{Type: t, Name: fi.Name})
+	})
 }
 
 // IsNotExist returns true if the error is caused by a not existing file.
@@ -307,7 +298,7 @@ func (be *beSwift) IsNotExist(err error) bool {
 // It will not remove the container itself.
 func (be *beSwift) Delete(ctx context.Context) error {
 	alltypes := []restic.FileType{
-		restic.DataFile,
+		restic.PackFile,
 		restic.KeyFile,
 		restic.LockFile,
 		restic.SnapshotFile,
