@@ -10,7 +10,6 @@ import (
 	"os"
 	"reflect"
 	"sort"
-	"strings"
 	"testing"
 	"time"
 
@@ -80,12 +79,12 @@ func (s *Suite) TestConfig(t *testing.T) {
 	var testString = "Config"
 
 	// create config and read it back
-	_, err := backend.LoadAll(context.TODO(), b, restic.Handle{Type: restic.ConfigFile})
+	_, err := backend.LoadAll(context.TODO(), nil, b, restic.Handle{Type: restic.ConfigFile})
 	if err == nil {
 		t.Fatalf("did not get expected error for non-existing config")
 	}
 
-	err = b.Save(context.TODO(), restic.Handle{Type: restic.ConfigFile}, strings.NewReader(testString))
+	err = b.Save(context.TODO(), restic.Handle{Type: restic.ConfigFile}, restic.NewByteReader([]byte(testString)))
 	if err != nil {
 		t.Fatalf("Save() error: %+v", err)
 	}
@@ -94,7 +93,7 @@ func (s *Suite) TestConfig(t *testing.T) {
 	// same config
 	for _, name := range []string{"", "foo", "bar", "0000000000000000000000000000000000000000000000000000000000000000"} {
 		h := restic.Handle{Type: restic.ConfigFile, Name: name}
-		buf, err := backend.LoadAll(context.TODO(), b, h)
+		buf, err := backend.LoadAll(context.TODO(), nil, b, h)
 		if err != nil {
 			t.Fatalf("unable to read config with name %q: %+v", name, err)
 		}
@@ -115,15 +114,16 @@ func (s *Suite) TestLoad(t *testing.T) {
 	b := s.open(t)
 	defer s.close(t, b)
 
-	rd, err := b.Load(context.TODO(), restic.Handle{}, 0, 0)
+	noop := func(rd io.Reader) error {
+		return nil
+	}
+
+	err := b.Load(context.TODO(), restic.Handle{}, 0, 0, noop)
 	if err == nil {
 		t.Fatalf("Load() did not return an error for invalid handle")
 	}
-	if rd != nil {
-		_ = rd.Close()
-	}
 
-	err = testLoad(b, restic.Handle{Type: restic.DataFile, Name: "foobar"}, 0, 0)
+	err = testLoad(b, restic.Handle{Type: restic.PackFile, Name: "foobar"}, 0, 0)
 	if err == nil {
 		t.Fatalf("Load() did not return an error for non-existing blob")
 	}
@@ -133,21 +133,31 @@ func (s *Suite) TestLoad(t *testing.T) {
 	data := test.Random(23, length)
 	id := restic.Hash(data)
 
-	handle := restic.Handle{Type: restic.DataFile, Name: id.String()}
-	err = b.Save(context.TODO(), handle, bytes.NewReader(data))
+	handle := restic.Handle{Type: restic.PackFile, Name: id.String()}
+	err = b.Save(context.TODO(), handle, restic.NewByteReader(data))
 	if err != nil {
 		t.Fatalf("Save() error: %+v", err)
 	}
 
 	t.Logf("saved %d bytes as %v", length, handle)
 
-	rd, err = b.Load(context.TODO(), handle, 100, -1)
+	err = b.Load(context.TODO(), handle, 100, -1, noop)
 	if err == nil {
 		t.Fatalf("Load() returned no error for negative offset!")
 	}
 
-	if rd != nil {
-		t.Fatalf("Load() returned a non-nil reader for negative offset!")
+	err = b.Load(context.TODO(), handle, 0, 0, func(rd io.Reader) error {
+		_, err := io.Copy(ioutil.Discard, rd)
+		if err != nil {
+			t.Fatal(err)
+		}
+		return errors.Errorf("deliberate error")
+	})
+	if err == nil {
+		t.Fatalf("Load() did not propagate consumer error!")
+	}
+	if err.Error() != "deliberate error" {
+		t.Fatalf("Load() did not correctly propagate consumer error!")
 	}
 
 	loadTests := 50
@@ -176,63 +186,38 @@ func (s *Suite) TestLoad(t *testing.T) {
 			d = d[:l]
 		}
 
-		rd, err := b.Load(context.TODO(), handle, getlen, int64(o))
+		var buf []byte
+		err := b.Load(context.TODO(), handle, getlen, int64(o), func(rd io.Reader) (ierr error) {
+			buf, ierr = ioutil.ReadAll(rd)
+			return ierr
+		})
 		if err != nil {
 			t.Logf("Load, l %v, o %v, len(d) %v, getlen %v", l, o, len(d), getlen)
 			t.Errorf("Load(%d, %d) returned unexpected error: %+v", l, o, err)
 			continue
 		}
 
-		buf, err := ioutil.ReadAll(rd)
-		if err != nil {
-			t.Logf("Load, l %v, o %v, len(d) %v, getlen %v", l, o, len(d), getlen)
-			t.Errorf("Load(%d, %d) ReadAll() returned unexpected error: %+v", l, o, err)
-			if err = rd.Close(); err != nil {
-				t.Errorf("Load(%d, %d) rd.Close() returned error: %+v", l, o, err)
-			}
-			continue
-		}
-
 		if l == 0 && len(buf) != len(d) {
 			t.Logf("Load, l %v, o %v, len(d) %v, getlen %v", l, o, len(d), getlen)
 			t.Errorf("Load(%d, %d) wrong number of bytes read: want %d, got %d", l, o, len(d), len(buf))
-			if err = rd.Close(); err != nil {
-				t.Errorf("Load(%d, %d) rd.Close() returned error: %+v", l, o, err)
-			}
 			continue
 		}
 
 		if l > 0 && l <= len(d) && len(buf) != l {
 			t.Logf("Load, l %v, o %v, len(d) %v, getlen %v", l, o, len(d), getlen)
 			t.Errorf("Load(%d, %d) wrong number of bytes read: want %d, got %d", l, o, l, len(buf))
-			if err = rd.Close(); err != nil {
-				t.Errorf("Load(%d, %d) rd.Close() returned error: %+v", l, o, err)
-			}
 			continue
 		}
 
 		if l > len(d) && len(buf) != len(d) {
 			t.Logf("Load, l %v, o %v, len(d) %v, getlen %v", l, o, len(d), getlen)
 			t.Errorf("Load(%d, %d) wrong number of bytes read for overlong read: want %d, got %d", l, o, l, len(buf))
-			if err = rd.Close(); err != nil {
-				t.Errorf("Load(%d, %d) rd.Close() returned error: %+v", l, o, err)
-			}
 			continue
 		}
 
 		if !bytes.Equal(buf, d) {
 			t.Logf("Load, l %v, o %v, len(d) %v, getlen %v", l, o, len(d), getlen)
 			t.Errorf("Load(%d, %d) returned wrong bytes", l, o)
-			if err = rd.Close(); err != nil {
-				t.Errorf("Load(%d, %d) rd.Close() returned error: %+v", l, o, err)
-			}
-			continue
-		}
-
-		err = rd.Close()
-		if err != nil {
-			t.Logf("Load, l %v, o %v, len(d) %v, getlen %v", l, o, len(d), getlen)
-			t.Errorf("Load(%d, %d) rd.Close() returned unexpected error: %+v", l, o, err)
 			continue
 		}
 	}
@@ -249,17 +234,30 @@ func (s *Suite) TestList(t *testing.T) {
 	b := s.open(t)
 	defer s.close(t, b)
 
-	list1 := restic.NewIDSet()
+	// Check that the backend is empty to start with
+	var found []string
+	err := b.List(context.TODO(), restic.PackFile, func(fi restic.FileInfo) error {
+		found = append(found, fi.Name)
+		return nil
+	})
+	if err != nil {
+		t.Fatalf("List returned error %v", err)
+	}
+	if found != nil {
+		t.Fatalf("backend not empty at start of test - contains: %v", found)
+	}
+
+	list1 := make(map[restic.ID]int64)
 
 	for i := 0; i < numTestFiles; i++ {
-		data := []byte(fmt.Sprintf("random test blob %v", i))
+		data := test.Random(rand.Int(), rand.Intn(100)+55)
 		id := restic.Hash(data)
-		h := restic.Handle{Type: restic.DataFile, Name: id.String()}
-		err := b.Save(context.TODO(), h, bytes.NewReader(data))
+		h := restic.Handle{Type: restic.PackFile, Name: id.String()}
+		err := b.Save(context.TODO(), h, restic.NewByteReader(data))
 		if err != nil {
 			t.Fatal(err)
 		}
-		list1.Insert(id)
+		list1[id] = int64(len(data))
 	}
 
 	t.Logf("wrote %v files", len(list1))
@@ -272,7 +270,7 @@ func (s *Suite) TestList(t *testing.T) {
 
 	for _, test := range tests {
 		t.Run(fmt.Sprintf("max-%v", test.maxItems), func(t *testing.T) {
-			list2 := restic.NewIDSet()
+			list2 := make(map[restic.ID]int64)
 
 			type setter interface {
 				SetListMaxItems(int)
@@ -283,19 +281,37 @@ func (s *Suite) TestList(t *testing.T) {
 				s.SetListMaxItems(test.maxItems)
 			}
 
-			for name := range b.List(context.TODO(), restic.DataFile) {
-				id, err := restic.ParseID(name)
+			err := b.List(context.TODO(), restic.PackFile, func(fi restic.FileInfo) error {
+				id, err := restic.ParseID(fi.Name)
 				if err != nil {
 					t.Fatal(err)
 				}
-				list2.Insert(id)
+				list2[id] = fi.Size
+				return nil
+			})
+
+			if err != nil {
+				t.Fatalf("List returned error %v", err)
 			}
 
 			t.Logf("loaded %v IDs from backend", len(list2))
 
-			if !list1.Equals(list2) {
-				t.Errorf("lists are not equal, list1 %d entries, list2 %d entries",
-					len(list1), len(list2))
+			for id, size := range list1 {
+				size2, ok := list2[id]
+				if !ok {
+					t.Errorf("id %v not returned by List()", id.Str())
+				}
+
+				if size != size2 {
+					t.Errorf("wrong size for id %v returned: want %v, got %v", id.Str(), size, size2)
+				}
+			}
+
+			for id := range list2 {
+				_, ok := list1[id]
+				if !ok {
+					t.Errorf("extra id %v returned by List()", id.Str())
+				}
 			}
 		})
 	}
@@ -303,18 +319,133 @@ func (s *Suite) TestList(t *testing.T) {
 	t.Logf("remove %d files", numTestFiles)
 	handles := make([]restic.Handle, 0, len(list1))
 	for id := range list1 {
-		handles = append(handles, restic.Handle{Type: restic.DataFile, Name: id.String()})
+		handles = append(handles, restic.Handle{Type: restic.PackFile, Name: id.String()})
 	}
 
-	err := s.delayedRemove(t, b, handles...)
+	err = s.delayedRemove(t, b, handles...)
+	if err != nil {
+		t.Fatal(err)
+	}
+}
+
+// TestListCancel tests that the context is respected and the error is returned by List.
+func (s *Suite) TestListCancel(t *testing.T) {
+	seedRand(t)
+
+	numTestFiles := 5
+
+	b := s.open(t)
+	defer s.close(t, b)
+
+	testFiles := make([]restic.Handle, 0, numTestFiles)
+
+	for i := 0; i < numTestFiles; i++ {
+		data := []byte(fmt.Sprintf("random test blob %v", i))
+		id := restic.Hash(data)
+		h := restic.Handle{Type: restic.PackFile, Name: id.String()}
+		err := b.Save(context.TODO(), h, restic.NewByteReader(data))
+		if err != nil {
+			t.Fatal(err)
+		}
+		testFiles = append(testFiles, h)
+	}
+
+	t.Run("Cancelled", func(t *testing.T) {
+		ctx, cancel := context.WithCancel(context.TODO())
+		cancel()
+
+		// pass in a cancelled context
+		err := b.List(ctx, restic.PackFile, func(fi restic.FileInfo) error {
+			t.Errorf("got FileInfo %v for cancelled context", fi)
+			return nil
+		})
+
+		if errors.Cause(err) != context.Canceled {
+			t.Fatalf("expected error not found, want %v, got %v", context.Canceled, errors.Cause(err))
+		}
+	})
+
+	t.Run("First", func(t *testing.T) {
+		ctx, cancel := context.WithCancel(context.TODO())
+		defer cancel()
+
+		i := 0
+		err := b.List(ctx, restic.PackFile, func(fi restic.FileInfo) error {
+			i++
+			// cancel the context on the first file
+			if i == 1 {
+				cancel()
+			}
+			return nil
+		})
+
+		if errors.Cause(err) != context.Canceled {
+			t.Fatalf("expected error not found, want %v, got %v", context.Canceled, err)
+		}
+
+		if i != 1 {
+			t.Fatalf("wrong number of files returned by List, want %v, got %v", 1, i)
+		}
+	})
+
+	t.Run("Last", func(t *testing.T) {
+		ctx, cancel := context.WithCancel(context.TODO())
+		defer cancel()
+
+		i := 0
+		err := b.List(ctx, restic.PackFile, func(fi restic.FileInfo) error {
+			// cancel the context at the last file
+			i++
+			if i == numTestFiles {
+				cancel()
+			}
+			return nil
+		})
+
+		if errors.Cause(err) != context.Canceled {
+			t.Fatalf("expected error not found, want %v, got %v", context.Canceled, err)
+		}
+
+		if i != numTestFiles {
+			t.Fatalf("wrong number of files returned by List, want %v, got %v", numTestFiles, i)
+		}
+	})
+
+	t.Run("Timeout", func(t *testing.T) {
+		// rather large timeout, let's try to get at least one item
+		timeout := time.Second
+
+		ctxTimeout, cancel := context.WithTimeout(context.TODO(), timeout)
+		defer cancel()
+
+		i := 0
+		// pass in a context with a timeout
+		err := b.List(ctxTimeout, restic.PackFile, func(fi restic.FileInfo) error {
+			i++
+
+			// wait until the context is cancelled
+			<-ctxTimeout.Done()
+			return nil
+		})
+
+		if errors.Cause(err) != context.DeadlineExceeded {
+			t.Fatalf("expected error not found, want %#v, got %#v", context.DeadlineExceeded, err)
+		}
+
+		if i > 2 {
+			t.Fatalf("wrong number of files returned by List, want <= 2, got %v", i)
+		}
+	})
+
+	err := s.delayedRemove(t, b, testFiles...)
 	if err != nil {
 		t.Fatal(err)
 	}
 }
 
 type errorCloser struct {
-	io.Reader
-	l int
+	io.ReadSeeker
+	l int64
 	t testing.TB
 }
 
@@ -323,10 +454,15 @@ func (ec errorCloser) Close() error {
 	return errors.New("forbidden method close was called")
 }
 
-func (ec errorCloser) Len() int {
+func (ec errorCloser) Length() int64 {
 	return ec.l
 }
 
+func (ec errorCloser) Rewind() error {
+	_, err := ec.ReadSeeker.Seek(0, io.SeekStart)
+	return err
+}
+
 // TestSave tests saving data in the backend.
 func (s *Suite) TestSave(t *testing.T) {
 	seedRand(t)
@@ -347,13 +483,13 @@ func (s *Suite) TestSave(t *testing.T) {
 		copy(id[:], data)
 
 		h := restic.Handle{
-			Type: restic.DataFile,
+			Type: restic.PackFile,
 			Name: fmt.Sprintf("%s-%d", id, i),
 		}
-		err := b.Save(context.TODO(), h, bytes.NewReader(data))
+		err := b.Save(context.TODO(), h, restic.NewByteReader(data))
 		test.OK(t, err)
 
-		buf, err := backend.LoadAll(context.TODO(), b, h)
+		buf, err := backend.LoadAll(context.TODO(), nil, b, h)
 		test.OK(t, err)
 		if len(buf) != len(data) {
 			t.Fatalf("number of bytes does not match, want %v, got %v", len(data), len(buf))
@@ -366,8 +502,12 @@ func (s *Suite) TestSave(t *testing.T) {
 		fi, err := b.Stat(context.TODO(), h)
 		test.OK(t, err)
 
+		if fi.Name != h.Name {
+			t.Errorf("Stat() returned wrong name, want %q, got %q", h.Name, fi.Name)
+		}
+
 		if fi.Size != int64(len(data)) {
-			t.Fatalf("Stat() returned different size, want %q, got %d", len(data), fi.Size)
+			t.Errorf("Stat() returned different size, want %q, got %d", len(data), fi.Size)
 		}
 
 		err = b.Remove(context.TODO(), h)
@@ -394,11 +534,11 @@ func (s *Suite) TestSave(t *testing.T) {
 		t.Fatal(err)
 	}
 
-	h := restic.Handle{Type: restic.DataFile, Name: id.String()}
+	h := restic.Handle{Type: restic.PackFile, Name: id.String()}
 
 	// wrap the tempfile in an errorCloser, so we can detect if the backend
 	// closes the reader
-	err = b.Save(context.TODO(), h, errorCloser{t: t, l: length, Reader: tmpfile})
+	err = b.Save(context.TODO(), h, errorCloser{t: t, l: int64(length), ReadSeeker: tmpfile})
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -408,25 +548,10 @@ func (s *Suite) TestSave(t *testing.T) {
 		t.Fatalf("error removing item: %+v", err)
 	}
 
-	// try again directly with the temp file
-	if _, err = tmpfile.Seek(588, io.SeekStart); err != nil {
-		t.Fatal(err)
-	}
-
-	err = b.Save(context.TODO(), h, tmpfile)
-	if err != nil {
-		t.Fatal(err)
-	}
-
 	if err = tmpfile.Close(); err != nil {
 		t.Fatal(err)
 	}
 
-	err = b.Remove(context.TODO(), h)
-	if err != nil {
-		t.Fatalf("error removing item: %+v", err)
-	}
-
 	if err = os.Remove(tmpfile.Name()); err != nil {
 		t.Fatal(err)
 	}
@@ -450,14 +575,14 @@ func (s *Suite) TestSaveFilenames(t *testing.T) {
 	defer s.close(t, b)
 
 	for i, test := range filenameTests {
-		h := restic.Handle{Name: test.name, Type: restic.DataFile}
-		err := b.Save(context.TODO(), h, strings.NewReader(test.data))
+		h := restic.Handle{Name: test.name, Type: restic.PackFile}
+		err := b.Save(context.TODO(), h, restic.NewByteReader([]byte(test.data)))
 		if err != nil {
 			t.Errorf("test %d failed: Save() returned %+v", i, err)
 			continue
 		}
 
-		buf, err := backend.LoadAll(context.TODO(), b, h)
+		buf, err := backend.LoadAll(context.TODO(), nil, b, h)
 		if err != nil {
 			t.Errorf("test %d failed: Load() returned %+v", i, err)
 			continue
@@ -488,24 +613,17 @@ var testStrings = []struct {
 func store(t testing.TB, b restic.Backend, tpe restic.FileType, data []byte) restic.Handle {
 	id := restic.Hash(data)
 	h := restic.Handle{Name: id.String(), Type: tpe}
-	err := b.Save(context.TODO(), h, bytes.NewReader(data))
+	err := b.Save(context.TODO(), h, restic.NewByteReader([]byte(data)))
 	test.OK(t, err)
 	return h
 }
 
 // testLoad loads a blob (but discards its contents).
 func testLoad(b restic.Backend, h restic.Handle, length int, offset int64) error {
-	rd, err := b.Load(context.TODO(), h, 0, 0)
-	if err != nil {
-		return err
-	}
-
-	_, err = io.Copy(ioutil.Discard, rd)
-	cerr := rd.Close()
-	if err == nil {
-		err = cerr
-	}
-	return err
+	return b.Load(context.TODO(), h, 0, 0, func(rd io.Reader) (ierr error) {
+		_, ierr = io.Copy(ioutil.Discard, rd)
+		return ierr
+	})
 }
 
 func (s *Suite) delayedRemove(t testing.TB, be restic.Backend, handles ...restic.Handle) error {
@@ -556,10 +674,16 @@ func delayedList(t testing.TB, b restic.Backend, tpe restic.FileType, max int, m
 	list := restic.NewIDSet()
 	start := time.Now()
 	for i := 0; i < max; i++ {
-		for s := range b.List(context.TODO(), tpe) {
-			id := restic.TestParseID(s)
+		err := b.List(context.TODO(), tpe, func(fi restic.FileInfo) error {
+			id := restic.TestParseID(fi.Name)
 			list.Insert(id)
+			return nil
+		})
+
+		if err != nil {
+			t.Fatal(err)
 		}
+
 		if len(list) < max && time.Since(start) < maxwait {
 			time.Sleep(500 * time.Millisecond)
 		}
@@ -574,7 +698,7 @@ func (s *Suite) TestBackend(t *testing.T) {
 	defer s.close(t, b)
 
 	for _, tpe := range []restic.FileType{
-		restic.DataFile, restic.KeyFile, restic.LockFile,
+		restic.PackFile, restic.KeyFile, restic.LockFile,
 		restic.SnapshotFile, restic.IndexFile,
 	} {
 		// detect non-existing files
@@ -608,7 +732,7 @@ func (s *Suite) TestBackend(t *testing.T) {
 
 			// test Load()
 			h := restic.Handle{Type: tpe, Name: ts.id}
-			buf, err := backend.LoadAll(context.TODO(), b, h)
+			buf, err := backend.LoadAll(context.TODO(), nil, b, h)
 			test.OK(t, err)
 			test.Equals(t, ts.data, string(buf))
 
@@ -618,31 +742,23 @@ func (s *Suite) TestBackend(t *testing.T) {
 			length := end - start
 
 			buf2 := make([]byte, length)
-			rd, err := b.Load(context.TODO(), h, len(buf2), int64(start))
+			var n int
+			err = b.Load(context.TODO(), h, len(buf2), int64(start), func(rd io.Reader) (ierr error) {
+				n, ierr = io.ReadFull(rd, buf2)
+				return ierr
+			})
 			test.OK(t, err)
-			n, err := io.ReadFull(rd, buf2)
 			test.OK(t, err)
 			test.Equals(t, len(buf2), n)
-
-			remaining, err := io.Copy(ioutil.Discard, rd)
-			test.OK(t, err)
-			test.Equals(t, int64(0), remaining)
-
-			test.OK(t, rd.Close())
-
 			test.Equals(t, ts.data[start:end], string(buf2))
 		}
 
 		// test adding the first file again
 		ts := testStrings[0]
-
-		// create blob
 		h := restic.Handle{Type: tpe, Name: ts.id}
-		err := b.Save(context.TODO(), h, strings.NewReader(ts.data))
-		test.Assert(t, err != nil, "expected error for %v, got %v", h, err)
 
 		// remove and recreate
-		err = s.delayedRemove(t, b, h)
+		err := s.delayedRemove(t, b, h)
 		test.OK(t, err)
 
 		// test that the blob is gone
@@ -651,7 +767,7 @@ func (s *Suite) TestBackend(t *testing.T) {
 		test.Assert(t, !ok, "removed blob still present")
 
 		// create blob
-		err = b.Save(context.TODO(), h, strings.NewReader(ts.data))
+		err = b.Save(context.TODO(), h, restic.NewByteReader([]byte(ts.data)))
 		test.OK(t, err)
 
 		// list items
