@@ -5,7 +5,6 @@ import (
 	"io"
 	"sync"
 
-	"github.com/pkg/errors"
 	"github.com/restic/restic/internal/debug"
 	"github.com/restic/restic/internal/restic"
 )
@@ -41,50 +40,44 @@ func (b *Backend) Remove(ctx context.Context, h restic.Handle) error {
 		return err
 	}
 
-	return b.Cache.Remove(h)
+	return b.Cache.remove(h)
 }
 
 var autoCacheTypes = map[restic.FileType]struct{}{
-	restic.IndexFile:    struct{}{},
-	restic.SnapshotFile: struct{}{},
+	restic.IndexFile:    {},
+	restic.SnapshotFile: {},
 }
 
 // Save stores a new file in the backend and the cache.
-func (b *Backend) Save(ctx context.Context, h restic.Handle, rd io.Reader) (err error) {
+func (b *Backend) Save(ctx context.Context, h restic.Handle, rd restic.RewindReader) error {
 	if _, ok := autoCacheTypes[h.Type]; !ok {
 		return b.Backend.Save(ctx, h, rd)
 	}
 
 	debug.Log("Save(%v): auto-store in the cache", h)
 
-	seeker, ok := rd.(io.Seeker)
-	if !ok {
-		return errors.New("reader is not a seeker")
-	}
-
-	pos, err := seeker.Seek(0, io.SeekCurrent)
+	// make sure the reader is at the start
+	err := rd.Rewind()
 	if err != nil {
-		return errors.Wrapf(err, "Seek")
-	}
-
-	if pos != 0 {
-		return errors.Errorf("reader is not rewind (pos %d)", pos)
+		return err
 	}
 
+	// first, save in the backend
 	err = b.Backend.Save(ctx, h, rd)
 	if err != nil {
 		return err
 	}
 
-	_, err = seeker.Seek(pos, io.SeekStart)
+	// next, save in the cache
+	err = rd.Rewind()
 	if err != nil {
-		return errors.Wrapf(err, "Seek")
+		return err
 	}
 
 	err = b.Cache.Save(h, rd)
 	if err != nil {
 		debug.Log("unable to save %v to cache: %v", h, err)
-		_ = b.Cache.Remove(h)
+		_ = b.Cache.remove(h)
 		return nil
 	}
 
@@ -98,14 +91,6 @@ var autoCacheFiles = map[restic.FileType]bool{
 
 func (b *Backend) cacheFile(ctx context.Context, h restic.Handle) error {
 	finish := make(chan struct{})
-	defer func() {
-		close(finish)
-
-		// remove the finish channel from the map
-		b.inProgressMutex.Lock()
-		delete(b.inProgress, h)
-		b.inProgressMutex.Unlock()
-	}()
 
 	b.inProgressMutex.Lock()
 	other, alreadyDownloading := b.inProgress[h]
@@ -121,37 +106,50 @@ func (b *Backend) cacheFile(ctx context.Context, h restic.Handle) error {
 		return nil
 	}
 
-	rd, err := b.Backend.Load(ctx, h, 0, 0)
-	if err != nil {
-		return err
+	// test again, maybe the file was cached in the meantime
+	if !b.Cache.Has(h) {
+
+		// nope, it's still not in the cache, pull it from the repo and save it
+
+		err := b.Backend.Load(ctx, h, 0, 0, func(rd io.Reader) error {
+			return b.Cache.Save(h, rd)
+		})
+		if err != nil {
+			// try to remove from the cache, ignore errors
+			_ = b.Cache.remove(h)
+		}
 	}
 
-	if err = b.Cache.Save(h, rd); err != nil {
-		return err
-	}
+	// signal other waiting goroutines that the file may now be cached
+	close(finish)
 
-	if err = rd.Close(); err != nil {
-		// try to remove from the cache, ignore errors
-		_ = b.Cache.Remove(h)
-		return err
-	}
+	// remove the finish channel from the map
+	b.inProgressMutex.Lock()
+	delete(b.inProgress, h)
+	b.inProgressMutex.Unlock()
 
 	return nil
 }
 
 // loadFromCacheOrDelegate will try to load the file from the cache, and fall
 // back to the backend if that fails.
-func (b *Backend) loadFromCacheOrDelegate(ctx context.Context, h restic.Handle, length int, offset int64) (io.ReadCloser, error) {
-	rd, err := b.Cache.Load(h, length, offset)
-	if err == nil {
-		return rd, nil
+func (b *Backend) loadFromCacheOrDelegate(ctx context.Context, h restic.Handle, length int, offset int64, consumer func(rd io.Reader) error) error {
+	rd, err := b.Cache.load(h, length, offset)
+	if err != nil {
+		debug.Log("error caching %v: %v, falling back to backend", h, err)
+		return b.Backend.Load(ctx, h, length, offset, consumer)
 	}
 
-	return b.Backend.Load(ctx, h, length, offset)
+	err = consumer(rd)
+	if err != nil {
+		_ = rd.Close() // ignore secondary errors
+		return err
+	}
+	return rd.Close()
 }
 
 // Load loads a file from the cache or the backend.
-func (b *Backend) Load(ctx context.Context, h restic.Handle, length int, offset int64) (io.ReadCloser, error) {
+func (b *Backend) Load(ctx context.Context, h restic.Handle, length int, offset int64, consumer func(rd io.Reader) error) error {
 	b.inProgressMutex.Lock()
 	waitForFinish, inProgress := b.inProgress[h]
 	b.inProgressMutex.Unlock()
@@ -164,9 +162,14 @@ func (b *Backend) Load(ctx context.Context, h restic.Handle, length int, offset
 
 	if b.Cache.Has(h) {
 		debug.Log("Load(%v, %v, %v) from cache", h, length, offset)
-		rd, err := b.Cache.Load(h, length, offset)
+		rd, err := b.Cache.load(h, length, offset)
 		if err == nil {
-			return rd, nil
+			err = consumer(rd)
+			if err != nil {
+				rd.Close() // ignore secondary errors
+				return err
+			}
+			return rd.Close()
 		}
 		debug.Log("error loading %v from cache: %v", h, err)
 	}
@@ -178,32 +181,30 @@ func (b *Backend) Load(ctx context.Context, h restic.Handle, length int, offset
 
 			err := b.cacheFile(ctx, h)
 			if err == nil {
-				return b.loadFromCacheOrDelegate(ctx, h, length, offset)
+				return b.loadFromCacheOrDelegate(ctx, h, length, offset, consumer)
 			}
 
 			debug.Log("error caching %v: %v", h, err)
 		}
 
 		debug.Log("Load(%v, %v, %v): partial file requested, delegating to backend", h, length, offset)
-		return b.Backend.Load(ctx, h, length, offset)
+		return b.Backend.Load(ctx, h, length, offset, consumer)
 	}
 
 	// if we don't automatically cache this file type, fall back to the backend
 	if _, ok := autoCacheFiles[h.Type]; !ok {
 		debug.Log("Load(%v, %v, %v): delegating to backend", h, length, offset)
-		return b.Backend.Load(ctx, h, length, offset)
+		return b.Backend.Load(ctx, h, length, offset, consumer)
 	}
 
 	debug.Log("auto-store %v in the cache", h)
 	err := b.cacheFile(ctx, h)
-
 	if err == nil {
-		// load the cached version
-		return b.Cache.Load(h, 0, 0)
+		return b.loadFromCacheOrDelegate(ctx, h, length, offset, consumer)
 	}
 
 	debug.Log("error caching %v: %v, falling back to backend", h, err)
-	return b.Backend.Load(ctx, h, length, offset)
+	return b.Backend.Load(ctx, h, length, offset, consumer)
 }
 
 // Stat tests whether the backend has a file. If it does not exist but still
@@ -215,7 +216,7 @@ func (b *Backend) Stat(ctx context.Context, h restic.Handle) (restic.FileInfo, e
 	if err != nil {
 		if b.Backend.IsNotExist(err) {
 			// try to remove from the cache, ignore errors
-			_ = b.Cache.Remove(h)
+			_ = b.Cache.remove(h)
 		}
 
 		return fi, err
