@@ -2,18 +2,12 @@ package repository
 
 import (
 	"context"
-	"crypto/sha256"
-	"fmt"
-	"io"
-	"os"
 
 	"github.com/restic/restic/internal/debug"
+	"github.com/restic/restic/internal/errors"
 	"github.com/restic/restic/internal/fs"
-	"github.com/restic/restic/internal/hashing"
 	"github.com/restic/restic/internal/pack"
 	"github.com/restic/restic/internal/restic"
-
-	"github.com/restic/restic/internal/errors"
 )
 
 // Repack takes a list of packs together with a list of blobs contained in
@@ -21,34 +15,23 @@ import (
 // into a new pack. Returned is the list of obsolete packs which can then
 // be removed.
 func Repack(ctx context.Context, repo restic.Repository, packs restic.IDSet, keepBlobs restic.BlobSet, p *restic.Progress) (obsoletePacks restic.IDSet, err error) {
+	if p != nil {
+		p.Start()
+		defer p.Done()
+	}
+
 	debug.Log("repacking %d packs while keeping %d blobs", len(packs), len(keepBlobs))
 
 	for packID := range packs {
 		// load the complete pack into a temp file
-		h := restic.Handle{Type: restic.DataFile, Name: packID.String()}
+		h := restic.Handle{Type: restic.PackFile, Name: packID.String()}
 
-		tempfile, err := fs.TempFile("", "restic-temp-repack-")
+		tempfile, hash, packLength, err := DownloadAndHash(ctx, repo.Backend(), h)
 		if err != nil {
-			return nil, errors.Wrap(err, "TempFile")
+			return nil, errors.Wrap(err, "Repack")
 		}
 
-		beRd, err := repo.Backend().Load(ctx, h, 0, 0)
-		if err != nil {
-			return nil, err
-		}
-
-		hrd := hashing.NewReader(beRd, sha256.New())
-		packLength, err := io.Copy(tempfile, hrd)
-		if err != nil {
-			return nil, errors.Wrap(err, "Copy")
-		}
-
-		if err = beRd.Close(); err != nil {
-			return nil, errors.Wrap(err, "Close")
-		}
-
-		hash := restic.IDFromHash(hrd.Sum(nil))
-		debug.Log("pack %v loaded (%d bytes), hash %v", packID.Str(), packLength, hash.Str())
+		debug.Log("pack %v loaded (%d bytes), hash %v", packID, packLength, hash)
 
 		if !packID.Equal(hash) {
 			return nil, errors.Errorf("hash does not match id: want %v, got %v", packID, hash)
@@ -64,7 +47,7 @@ func Repack(ctx context.Context, repo restic.Repository, packs restic.IDSet, kee
 			return nil, err
 		}
 
-		debug.Log("processing pack %v, blobs: %v", packID.Str(), len(blobs))
+		debug.Log("processing pack %v, blobs: %v", packID, len(blobs))
 		var buf []byte
 		for _, entry := range blobs {
 			h := restic.BlobHandle{ID: entry.ID, Type: entry.Type}
@@ -74,8 +57,7 @@ func Repack(ctx context.Context, repo restic.Repository, packs restic.IDSet, kee
 
 			debug.Log("  process blob %v", h)
 
-			buf = buf[:]
-			if uint(len(buf)) < entry.Length {
+			if uint(cap(buf)) < entry.Length {
 				buf = make([]byte, entry.Length)
 			}
 			buf = buf[:entry.Length]
@@ -100,16 +82,17 @@ func Repack(ctx context.Context, repo restic.Repository, packs restic.IDSet, kee
 			if !id.Equal(entry.ID) {
 				debug.Log("read blob %v/%v from %v: wrong data returned, hash is %v",
 					h.Type, h.ID, tempfile.Name(), id)
-				fmt.Fprintf(os.Stderr, "read blob %v from %v: wrong data returned, hash is %v",
+				return nil, errors.Errorf("read blob %v from %v: wrong data returned, hash is %v",
 					h, tempfile.Name(), id)
 			}
 
-			_, err = repo.SaveBlob(ctx, entry.Type, plaintext, entry.ID)
+			// We do want to save already saved blobs!
+			_, _, err = repo.SaveBlob(ctx, entry.Type, plaintext, entry.ID, true)
 			if err != nil {
 				return nil, err
 			}
 
-			debug.Log("  saved blob %v", entry.ID.Str())
+			debug.Log("  saved blob %v", entry.ID)
 
 			keepBlobs.Delete(h)
 		}
