@@ -5,16 +5,21 @@ import (
 	"context"
 	"encoding/json"
 	"fmt"
+	"io"
 	"os"
 
+	"github.com/restic/chunker"
 	"github.com/restic/restic/internal/cache"
-	"github.com/restic/restic/internal/errors"
-	"github.com/restic/restic/internal/restic"
-
-	"github.com/restic/restic/internal/backend"
 	"github.com/restic/restic/internal/crypto"
 	"github.com/restic/restic/internal/debug"
+	"github.com/restic/restic/internal/errors"
+	"github.com/restic/restic/internal/fs"
+	"github.com/restic/restic/internal/hashing"
 	"github.com/restic/restic/internal/pack"
+	"github.com/restic/restic/internal/restic"
+
+	"github.com/minio/sha256-simd"
+	"golang.org/x/sync/errgroup"
 )
 
 // Repository is used to access a repository in a backend.
@@ -24,7 +29,9 @@ type Repository struct {
 	key     *crypto.Key
 	keyName string
 	idx     *MasterIndex
-	restic.Cache
+	Cache   *cache.Cache
+
+	noAutoIndexUpdate bool
 
 	treePM *packerManager
 	dataPM *packerManager
@@ -42,13 +49,19 @@ func New(be restic.Backend) *Repository {
 	return repo
 }
 
+// DisableAutoIndexUpdate deactives the automatic finalization and upload of new
+// indexes once these are full
+func (r *Repository) DisableAutoIndexUpdate() {
+	r.noAutoIndexUpdate = true
+}
+
 // Config returns the repository configuration.
 func (r *Repository) Config() restic.Config {
 	return r.cfg
 }
 
 // UseCache replaces the backend with the wrapped cache.
-func (r *Repository) UseCache(c restic.Cache) {
+func (r *Repository) UseCache(c *cache.Cache) {
 	if c == nil {
 		return
 	}
@@ -59,19 +72,33 @@ func (r *Repository) UseCache(c restic.Cache) {
 
 // PrefixLength returns the number of bytes required so that all prefixes of
 // all IDs of type t are unique.
-func (r *Repository) PrefixLength(t restic.FileType) (int, error) {
-	return restic.PrefixLength(r.be, t)
+func (r *Repository) PrefixLength(ctx context.Context, t restic.FileType) (int, error) {
+	return restic.PrefixLength(ctx, r.be, t)
 }
 
-// LoadAndDecrypt loads and decrypts data identified by t and id from the
-// backend.
-func (r *Repository) LoadAndDecrypt(ctx context.Context, t restic.FileType, id restic.ID) (buf []byte, err error) {
-	debug.Log("load %v with id %v", t, id.Str())
+// LoadAndDecrypt loads and decrypts the file with the given type and ID, using
+// the supplied buffer (which must be empty). If the buffer is nil, a new
+// buffer will be allocated and returned.
+func (r *Repository) LoadAndDecrypt(ctx context.Context, buf []byte, t restic.FileType, id restic.ID) ([]byte, error) {
+	if len(buf) != 0 {
+		panic("buf is not empty")
+	}
+
+	debug.Log("load %v with id %v", t, id)
 
 	h := restic.Handle{Type: t, Name: id.String()}
-	buf, err = backend.LoadAll(ctx, r.be, h)
+	err := r.be.Load(ctx, h, 0, 0, func(rd io.Reader) error {
+		// make sure this call is idempotent, in case an error occurs
+		wr := bytes.NewBuffer(buf[:0])
+		_, cerr := io.Copy(wr, rd)
+		if cerr != nil {
+			return cerr
+		}
+		buf = wr.Bytes()
+		return nil
+	})
+
 	if err != nil {
-		debug.Log("error loading %v: %v", h, err)
 		return nil, err
 	}
 
@@ -88,60 +115,69 @@ func (r *Repository) LoadAndDecrypt(ctx context.Context, t restic.FileType, id r
 	return plaintext, nil
 }
 
-// sortCachedPacks moves all cached pack files to the front of blobs.
-func (r *Repository) sortCachedPacks(blobs []restic.PackedBlob) []restic.PackedBlob {
-	if r.Cache == nil {
-		return blobs
+type haver interface {
+	Has(restic.Handle) bool
+}
+
+// sortCachedPacksFirst moves all cached pack files to the front of blobs.
+func sortCachedPacksFirst(cache haver, blobs []restic.PackedBlob) {
+	if cache == nil {
+		return
 	}
 
-	cached := make([]restic.PackedBlob, 0, len(blobs)/2)
+	// no need to sort a list with one element
+	if len(blobs) == 1 {
+		return
+	}
+
+	cached := blobs[:0]
 	noncached := make([]restic.PackedBlob, 0, len(blobs)/2)
 
 	for _, blob := range blobs {
-		if r.Cache.Has(restic.Handle{Type: restic.DataFile, Name: blob.PackID.String()}) {
+		if cache.Has(restic.Handle{Type: restic.PackFile, Name: blob.PackID.String()}) {
 			cached = append(cached, blob)
 			continue
 		}
 		noncached = append(noncached, blob)
 	}
 
-	return append(cached, noncached...)
+	copy(blobs[len(cached):], noncached)
 }
 
-// loadBlob tries to load and decrypt content identified by t and id from a
-// pack from the backend, the result is stored in plaintextBuf, which must be
-// large enough to hold the complete blob.
-func (r *Repository) loadBlob(ctx context.Context, id restic.ID, t restic.BlobType, plaintextBuf []byte) (int, error) {
-	debug.Log("load %v with id %v (buf len %v, cap %d)", t, id.Str(), len(plaintextBuf), cap(plaintextBuf))
+// LoadBlob loads a blob of type t from the repository.
+// It may use all of buf[:cap(buf)] as scratch space.
+func (r *Repository) LoadBlob(ctx context.Context, t restic.BlobType, id restic.ID, buf []byte) ([]byte, error) {
+	debug.Log("load %v with id %v (buf len %v, cap %d)", t, id, len(buf), cap(buf))
 
 	// lookup packs
-	blobs, err := r.idx.Lookup(id, t)
-	if err != nil {
-		debug.Log("id %v not found in index: %v", id.Str(), err)
-		return 0, err
+	blobs := r.idx.Lookup(id, t)
+	if len(blobs) == 0 {
+		debug.Log("id %v not found in index", id)
+		return nil, errors.Errorf("id %v not found in repository", id)
 	}
 
 	// try cached pack files first
-	blobs = r.sortCachedPacks(blobs)
+	sortCachedPacksFirst(r.Cache, blobs)
 
 	var lastError error
 	for _, blob := range blobs {
-		debug.Log("blob %v/%v found: %v", t, id.Str(), blob)
+		debug.Log("blob %v/%v found: %v", t, id, blob)
 
 		if blob.Type != t {
 			debug.Log("blob %v has wrong block type, want %v", blob, t)
 		}
 
 		// load blob from pack
-		h := restic.Handle{Type: restic.DataFile, Name: blob.PackID.String()}
+		h := restic.Handle{Type: restic.PackFile, Name: blob.PackID.String()}
 
-		if uint(cap(plaintextBuf)) < blob.Length {
-			return 0, errors.Errorf("buffer is too small: %v < %v", cap(plaintextBuf), blob.Length)
+		switch {
+		case cap(buf) < int(blob.Length):
+			buf = make([]byte, blob.Length)
+		case len(buf) != int(blob.Length):
+			buf = buf[:blob.Length]
 		}
 
-		plaintextBuf = plaintextBuf[:blob.Length]
-
-		n, err := restic.ReadAt(ctx, r.be, h, int64(blob.Offset), plaintextBuf)
+		n, err := restic.ReadAt(ctx, r.be, h, int64(blob.Offset), buf)
 		if err != nil {
 			debug.Log("error loading blob %v: %v", blob, err)
 			lastError = err
@@ -156,7 +192,7 @@ func (r *Repository) loadBlob(ctx context.Context, id restic.ID, t restic.BlobTy
 		}
 
 		// decrypt
-		nonce, ciphertext := plaintextBuf[:r.key.NonceSize()], plaintextBuf[r.key.NonceSize():]
+		nonce, ciphertext := buf[:r.key.NonceSize()], buf[r.key.NonceSize():]
 		plaintext, err := r.key.Open(ciphertext[:0], nonce, ciphertext, nil)
 		if err != nil {
 			lastError = errors.Errorf("decrypting blob %v failed: %v", id, err)
@@ -169,22 +205,22 @@ func (r *Repository) loadBlob(ctx context.Context, id restic.ID, t restic.BlobTy
 			continue
 		}
 
-		// move decrypted data to the start of the provided buffer
-		copy(plaintextBuf[0:], plaintext)
-		return len(plaintext), nil
+		// move decrypted data to the start of the buffer
+		copy(buf, plaintext)
+		return buf[:len(plaintext)], nil
 	}
 
 	if lastError != nil {
-		return 0, lastError
+		return nil, lastError
 	}
 
-	return 0, errors.Errorf("loading blob %v from %v packs failed", id.Str(), len(blobs))
+	return nil, errors.Errorf("loading blob %v from %v packs failed", id.Str(), len(blobs))
 }
 
 // LoadJSONUnpacked decrypts the data and afterwards calls json.Unmarshal on
 // the item.
 func (r *Repository) LoadJSONUnpacked(ctx context.Context, t restic.FileType, id restic.ID, item interface{}) (err error) {
-	buf, err := r.LoadAndDecrypt(ctx, t, id)
+	buf, err := r.LoadAndDecrypt(ctx, nil, t, id)
 	if err != nil {
 		return err
 	}
@@ -193,27 +229,19 @@ func (r *Repository) LoadJSONUnpacked(ctx context.Context, t restic.FileType, id
 }
 
 // LookupBlobSize returns the size of blob id.
-func (r *Repository) LookupBlobSize(id restic.ID, tpe restic.BlobType) (uint, error) {
+func (r *Repository) LookupBlobSize(id restic.ID, tpe restic.BlobType) (uint, bool) {
 	return r.idx.LookupSize(id, tpe)
 }
 
 // SaveAndEncrypt encrypts data and stores it to the backend as type t. If data
 // is small enough, it will be packed together with other small blobs.
-func (r *Repository) SaveAndEncrypt(ctx context.Context, t restic.BlobType, data []byte, id *restic.ID) (restic.ID, error) {
-	if id == nil {
-		// compute plaintext hash
-		hashedID := restic.Hash(data)
-		id = &hashedID
-	}
+// The caller must ensure that the id matches the data.
+func (r *Repository) SaveAndEncrypt(ctx context.Context, t restic.BlobType, data []byte, id restic.ID) error {
+	debug.Log("save id %v (%v, %d bytes)", id, t, len(data))
 
-	debug.Log("save id %v (%v, %d bytes)", id.Str(), t, len(data))
-
-	// get buf from the pool
-	ciphertext := getBuf()
-	defer freeBuf(ciphertext)
-
-	ciphertext = ciphertext[:0]
 	nonce := crypto.NewRandomNonce()
+
+	ciphertext := make([]byte, 0, restic.CiphertextLength(len(data)))
 	ciphertext = append(ciphertext, nonce...)
 
 	// encrypt blob
@@ -233,24 +261,24 @@ func (r *Repository) SaveAndEncrypt(ctx context.Context, t restic.BlobType, data
 
 	packer, err := pm.findPacker()
 	if err != nil {
-		return restic.ID{}, err
+		return err
 	}
 
 	// save ciphertext
-	_, err = packer.Add(t, *id, ciphertext)
+	_, err = packer.Add(t, id, ciphertext)
 	if err != nil {
-		return restic.ID{}, err
+		return err
 	}
 
 	// if the pack is not full enough, put back to the list
 	if packer.Size() < minPackSize {
 		debug.Log("pack is not full enough (%d bytes)", packer.Size())
 		pm.insertPacker(packer)
-		return *id, nil
+		return nil
 	}
 
 	// else write the pack to the backend
-	return *id, r.savePacker(ctx, t, packer)
+	return r.savePacker(ctx, t, packer)
 }
 
 // SaveJSONUnpacked serialises item as JSON and encrypts and saves it in the
@@ -278,7 +306,7 @@ func (r *Repository) SaveUnpacked(ctx context.Context, t restic.FileType, p []by
 	id = restic.Hash(ciphertext)
 	h := restic.Handle{Type: t, Name: id.String()}
 
-	err = r.be.Save(ctx, h, bytes.NewReader(ciphertext))
+	err = r.be.Save(ctx, h, restic.NewByteReader(ciphertext))
 	if err != nil {
 		debug.Log("error saving blob %v: %v", h, err)
 		return restic.ID{}, err
@@ -288,8 +316,18 @@ func (r *Repository) SaveUnpacked(ctx context.Context, t restic.FileType, p []by
 	return id, nil
 }
 
-// Flush saves all remaining packs.
+// Flush saves all remaining packs and the index
 func (r *Repository) Flush(ctx context.Context) error {
+	if err := r.FlushPacks(ctx); err != nil {
+		return err
+	}
+
+	// Save index after flushing
+	return r.SaveIndex(ctx)
+}
+
+// FlushPacks saves all remaining packs.
+func (r *Repository) FlushPacks(ctx context.Context) error {
 	pms := []struct {
 		t  restic.BlobType
 		pm *packerManager
@@ -312,7 +350,6 @@ func (r *Repository) Flush(ctx context.Context) error {
 		p.pm.packers = p.pm.packers[:0]
 		p.pm.pm.Unlock()
 	}
-
 	return nil
 }
 
@@ -322,20 +359,34 @@ func (r *Repository) Backend() restic.Backend {
 }
 
 // Index returns the currently used MasterIndex.
-func (r *Repository) Index() restic.Index {
+func (r *Repository) Index() restic.MasterIndex {
 	return r.idx
 }
 
 // SetIndex instructs the repository to use the given index.
-func (r *Repository) SetIndex(i restic.Index) {
+func (r *Repository) SetIndex(i restic.MasterIndex) error {
 	r.idx = i.(*MasterIndex)
+
+	ids := restic.NewIDSet()
+	for _, idx := range r.idx.All() {
+		indexIDs, err := idx.IDs()
+		if err != nil {
+			debug.Log("not using index, ID() returned error %v", err)
+			continue
+		}
+		for _, id := range indexIDs {
+			ids.Insert(id)
+		}
+	}
+
+	return r.PrepareCache(ids)
 }
 
 // SaveIndex saves an index in the repository.
 func SaveIndex(ctx context.Context, repo restic.Repository, index *Index) (restic.ID, error) {
 	buf := bytes.NewBuffer(nil)
 
-	err := index.Finalize(buf)
+	err := index.Encode(buf)
 	if err != nil {
 		return restic.ID{}, err
 	}
@@ -353,131 +404,193 @@ func (r *Repository) saveIndex(ctx context.Context, indexes ...*Index) error {
 			return err
 		}
 
-		debug.Log("Saved index %d as %v", i, sid.Str())
+		debug.Log("Saved index %d as %v", i, sid)
 	}
+	r.idx.MergeFinalIndexes()
 
 	return nil
 }
 
 // SaveIndex saves all new indexes in the backend.
 func (r *Repository) SaveIndex(ctx context.Context) error {
-	return r.saveIndex(ctx, r.idx.NotFinalIndexes()...)
+	return r.saveIndex(ctx, r.idx.FinalizeNotFinalIndexes()...)
 }
 
 // SaveFullIndex saves all full indexes in the backend.
 func (r *Repository) SaveFullIndex(ctx context.Context) error {
-	return r.saveIndex(ctx, r.idx.FullIndexes()...)
+	return r.saveIndex(ctx, r.idx.FinalizeFullIndexes()...)
 }
 
-const loadIndexParallelism = 20
+const loadIndexParallelism = 4
 
 // LoadIndex loads all index files from the backend in parallel and stores them
 // in the master index. The first error that occurred is returned.
 func (r *Repository) LoadIndex(ctx context.Context) error {
 	debug.Log("Loading index")
 
-	errCh := make(chan error, 1)
-	indexes := make(chan *Index)
+	// track spawned goroutines using wg, create a new context which is
+	// cancelled as soon as an error occurs.
+	wg, ctx := errgroup.WithContext(ctx)
 
-	worker := func(ctx context.Context, id restic.ID) error {
-		idx, err := LoadIndex(ctx, r, id)
-		if err != nil {
-			fmt.Fprintf(os.Stderr, "%v, ignoring\n", err)
+	type FileInfo struct {
+		restic.ID
+		Size int64
+	}
+	ch := make(chan FileInfo)
+	indexCh := make(chan *Index)
+
+	// send list of index files through ch, which is closed afterwards
+	wg.Go(func() error {
+		defer close(ch)
+		return r.List(ctx, restic.IndexFile, func(id restic.ID, size int64) error {
+			select {
+			case <-ctx.Done():
+				return nil
+			case ch <- FileInfo{id, size}:
+			}
 			return nil
-		}
+		})
+	})
 
-		select {
-		case indexes <- idx:
-		case <-ctx.Done():
+	// a worker receives an index ID from ch, loads the index, and sends it to indexCh
+	worker := func() error {
+		var buf []byte
+		for fi := range ch {
+			var err error
+			buf, err = r.LoadAndDecrypt(ctx, buf[:0], restic.IndexFile, fi.ID)
+			if err != nil {
+				return errors.Wrapf(err, "unable to load index %s", fi.ID.Str())
+			}
+			idx, _, err := DecodeIndex(buf, fi.ID)
+			if err != nil {
+				return errors.Wrapf(err, "unable to decode index %s", fi.ID.Str())
+			}
+
+			select {
+			case indexCh <- idx:
+			case <-ctx.Done():
+			}
 		}
 
 		return nil
 	}
 
-	go func() {
-		defer close(indexes)
-		errCh <- FilesInParallel(ctx, r.be, restic.IndexFile, loadIndexParallelism,
-			ParallelWorkFuncParseID(worker))
-	}()
+	// run workers on ch
+	wg.Go(func() error {
+		defer close(indexCh)
+		return RunWorkers(loadIndexParallelism, worker)
+	})
 
+	// receive decoded indexes
 	validIndex := restic.NewIDSet()
-	for idx := range indexes {
-		id, err := idx.ID()
-		if err == nil {
-			validIndex.Insert(id)
+	wg.Go(func() error {
+		for idx := range indexCh {
+			ids, err := idx.IDs()
+			if err == nil {
+				for _, id := range ids {
+					validIndex.Insert(id)
+				}
+			}
+
+			r.idx.Insert(idx)
 		}
-		r.idx.Insert(idx)
+		r.idx.MergeFinalIndexes()
+		return nil
+	})
+
+	err := wg.Wait()
+	if err != nil {
+		return errors.Fatal(err.Error())
 	}
 
-	if r.Cache != nil {
-		// clear old index files
-		err := r.Cache.Clear(restic.IndexFile, validIndex)
-		if err != nil {
-			fmt.Fprintf(os.Stderr, "error clearing index files in cache: %v\n", err)
-		}
-
-		packs := restic.NewIDSet()
-		for _, idx := range r.idx.All() {
-			for id := range idx.Packs() {
-				packs.Insert(id)
-			}
-		}
-
-		// clear old data files
-		err = r.Cache.Clear(restic.DataFile, packs)
-		if err != nil {
-			fmt.Fprintf(os.Stderr, "error clearing data files in cache: %v\n", err)
-		}
-
-		treePacks := restic.NewIDSet()
-		for _, idx := range r.idx.All() {
-			for _, id := range idx.TreePacks() {
-				treePacks.Insert(id)
-			}
-		}
-
-		// use readahead
-		cache := r.Cache.(*cache.Cache)
-		cache.PerformReadahead = func(h restic.Handle) bool {
-			if h.Type != restic.DataFile {
-				return false
-			}
-
-			id, err := restic.ParseID(h.Name)
-			if err != nil {
-				return false
-			}
-
-			return treePacks.Has(id)
-		}
-	}
-
-	if err := <-errCh; err != nil {
+	// remove index files from the cache which have been removed in the repo
+	err = r.PrepareCache(validIndex)
+	if err != nil {
 		return err
 	}
 
 	return nil
 }
 
+// PrepareCache initializes the local cache. indexIDs is the list of IDs of
+// index files still present in the repo.
+func (r *Repository) PrepareCache(indexIDs restic.IDSet) error {
+	if r.Cache == nil {
+		return nil
+	}
+
+	debug.Log("prepare cache with %d index files", len(indexIDs))
+
+	// clear old index files
+	err := r.Cache.Clear(restic.IndexFile, indexIDs)
+	if err != nil {
+		fmt.Fprintf(os.Stderr, "error clearing index files in cache: %v\n", err)
+	}
+
+	packs := restic.NewIDSet()
+	for _, idx := range r.idx.All() {
+		for id := range idx.Packs() {
+			packs.Insert(id)
+		}
+	}
+
+	// clear old packs
+	err = r.Cache.Clear(restic.PackFile, packs)
+	if err != nil {
+		fmt.Fprintf(os.Stderr, "error clearing pack files in cache: %v\n", err)
+	}
+
+	treePacks := restic.NewIDSet()
+	for _, idx := range r.idx.All() {
+		for _, id := range idx.TreePacks() {
+			treePacks.Insert(id)
+		}
+	}
+
+	// use readahead
+	debug.Log("using readahead")
+	cache := r.Cache
+	cache.PerformReadahead = func(h restic.Handle) bool {
+		if h.Type != restic.PackFile {
+			debug.Log("no readahead for %v, is not a pack file", h)
+			return false
+		}
+
+		id, err := restic.ParseID(h.Name)
+		if err != nil {
+			debug.Log("no readahead for %v, invalid ID", h)
+			return false
+		}
+
+		if treePacks.Has(id) {
+			debug.Log("perform readahead for %v", h)
+			return true
+		}
+		debug.Log("no readahead for %v, not tree file", h)
+		return false
+	}
+
+	return nil
+}
+
 // LoadIndex loads the index id from backend and returns it.
 func LoadIndex(ctx context.Context, repo restic.Repository, id restic.ID) (*Index, error) {
-	idx, err := LoadIndexWithDecoder(ctx, repo, id, DecodeIndex)
-	if err == nil {
-		return idx, nil
+	buf, err := repo.LoadAndDecrypt(ctx, nil, restic.IndexFile, id)
+	if err != nil {
+		return nil, err
 	}
 
-	if errors.Cause(err) == ErrOldIndexFormat {
+	idx, oldFormat, err := DecodeIndex(buf, id)
+	if oldFormat {
 		fmt.Fprintf(os.Stderr, "index %v has old format\n", id.Str())
-		return LoadIndexWithDecoder(ctx, repo, id, DecodeOldIndex)
 	}
-
-	return nil, err
+	return idx, err
 }
 
 // SearchKey finds a key with the supplied password, afterwards the config is
 // read and parsed. It tries at most maxKeys key files in the repo.
-func (r *Repository) SearchKey(ctx context.Context, password string, maxKeys int) error {
-	key, err := SearchKey(ctx, r, password, maxKeys)
+func (r *Repository) SearchKey(ctx context.Context, password string, maxKeys int, keyHint string) error {
+	key, err := SearchKey(ctx, r, password, maxKeys, keyHint)
 	if err != nil {
 		return err
 	}
@@ -487,12 +600,15 @@ func (r *Repository) SearchKey(ctx context.Context, password string, maxKeys int
 	r.treePM.key = key.master
 	r.keyName = key.Name()
 	r.cfg, err = restic.LoadConfig(ctx, r)
-	return err
+	if err != nil {
+		return errors.Fatalf("config cannot be loaded: %v", err)
+	}
+	return nil
 }
 
 // Init creates a new master key with the supplied password, initializes and
 // saves the repository config.
-func (r *Repository) Init(ctx context.Context, password string) error {
+func (r *Repository) Init(ctx context.Context, password string, chunkerPolynomial *chunker.Pol) error {
 	has, err := r.be.Test(ctx, restic.Handle{Type: restic.ConfigFile})
 	if err != nil {
 		return err
@@ -505,6 +621,9 @@ func (r *Repository) Init(ctx context.Context, password string) error {
 	if err != nil {
 		return err
 	}
+	if chunkerPolynomial != nil {
+		cfg.ChunkerPolynomial = *chunkerPolynomial
+	}
 
 	return r.init(ctx, password, cfg)
 }
@@ -512,7 +631,7 @@ func (r *Repository) Init(ctx context.Context, password string) error {
 // init creates a new master key with the supplied password and uses it to save
 // the config into the repo.
 func (r *Repository) init(ctx context.Context, password string, cfg restic.Config) error {
-	key, err := createMasterKey(r, password)
+	key, err := createMasterKey(ctx, r, password)
 	if err != nil {
 		return err
 	}
@@ -536,40 +655,29 @@ func (r *Repository) KeyName() string {
 	return r.keyName
 }
 
-// List returns a channel that yields all IDs of type t in the backend.
-func (r *Repository) List(ctx context.Context, t restic.FileType) <-chan restic.ID {
-	out := make(chan restic.ID)
-	go func() {
-		defer close(out)
-		for strID := range r.be.List(ctx, t) {
-			if id, err := restic.ParseID(strID); err == nil {
-				select {
-				case out <- id:
-				case <-ctx.Done():
-					return
-				}
-			}
+// List runs fn for all files of type t in the repo.
+func (r *Repository) List(ctx context.Context, t restic.FileType, fn func(restic.ID, int64) error) error {
+	return r.be.List(ctx, t, func(fi restic.FileInfo) error {
+		id, err := restic.ParseID(fi.Name)
+		if err != nil {
+			debug.Log("unable to parse %v as an ID", fi.Name)
+			return nil
 		}
-	}()
-	return out
+		return fn(id, fi.Size)
+	})
 }
 
 // ListPack returns the list of blobs saved in the pack id and the length of
 // the file as stored in the backend.
-func (r *Repository) ListPack(ctx context.Context, id restic.ID) ([]restic.Blob, int64, error) {
-	h := restic.Handle{Type: restic.DataFile, Name: id.String()}
+func (r *Repository) ListPack(ctx context.Context, id restic.ID, size int64) ([]restic.Blob, int64, error) {
+	h := restic.Handle{Type: restic.PackFile, Name: id.String()}
 
-	blobInfo, err := r.Backend().Stat(ctx, h)
+	blobs, err := pack.List(r.Key(), restic.ReaderAt(ctx, r.Backend(), h), size)
 	if err != nil {
 		return nil, 0, err
 	}
 
-	blobs, err := pack.List(r.Key(), restic.ReaderAt(r.Backend(), h), blobInfo.Size)
-	if err != nil {
-		return nil, 0, err
-	}
-
-	return blobs, blobInfo.Size, nil
+	return blobs, size, nil
 }
 
 // Delete calls backend.Delete() if implemented, and returns an error
@@ -583,59 +691,40 @@ func (r *Repository) Close() error {
 	return r.be.Close()
 }
 
-// LoadBlob loads a blob of type t from the repository to the buffer. buf must
-// be large enough to hold the encrypted blob, since it is used as scratch
-// space.
-func (r *Repository) LoadBlob(ctx context.Context, t restic.BlobType, id restic.ID, buf []byte) (int, error) {
-	debug.Log("load blob %v into buf (len %v, cap %v)", id.Str(), len(buf), cap(buf))
-	size, err := r.idx.LookupSize(id, t)
-	if err != nil {
-		return 0, err
+// SaveBlob saves a blob of type t into the repository.
+// It takes care that no duplicates are saved; this can be overwritten
+// by setting storeDuplicate to true.
+// If id is the null id, it will be computed and returned.
+// Also returns if the blob was already known before
+func (r *Repository) SaveBlob(ctx context.Context, t restic.BlobType, buf []byte, id restic.ID, storeDuplicate bool) (newID restic.ID, known bool, err error) {
+
+	// compute plaintext hash if not already set
+	if id.IsNull() {
+		newID = restic.Hash(buf)
+	} else {
+		newID = id
 	}
 
-	if cap(buf) < restic.CiphertextLength(int(size)) {
-		return 0, errors.Errorf("buffer is too small for data blob (%d < %d)", cap(buf), restic.CiphertextLength(int(size)))
+	// first try to add to pending blobs; if not successful, this blob is already known
+	known = !r.idx.addPending(newID, t)
+
+	// only save when needed or explicitly told
+	if !known || storeDuplicate {
+		err = r.SaveAndEncrypt(ctx, t, buf, newID)
 	}
 
-	n, err := r.loadBlob(ctx, id, t, buf)
-	if err != nil {
-		return 0, err
-	}
-	buf = buf[:n]
-
-	debug.Log("loaded %d bytes into buf %p", len(buf), buf)
-
-	return len(buf), err
-}
-
-// SaveBlob saves a blob of type t into the repository. If id is the null id, it
-// will be computed and returned.
-func (r *Repository) SaveBlob(ctx context.Context, t restic.BlobType, buf []byte, id restic.ID) (restic.ID, error) {
-	var i *restic.ID
-	if !id.IsNull() {
-		i = &id
-	}
-	return r.SaveAndEncrypt(ctx, t, buf, i)
+	return newID, known, err
 }
 
 // LoadTree loads a tree from the repository.
 func (r *Repository) LoadTree(ctx context.Context, id restic.ID) (*restic.Tree, error) {
-	debug.Log("load tree %v", id.Str())
+	debug.Log("load tree %v", id)
 
-	size, err := r.idx.LookupSize(id, restic.TreeBlob)
+	buf, err := r.LoadBlob(ctx, restic.TreeBlob, id, nil)
 	if err != nil {
 		return nil, err
 	}
 
-	debug.Log("size is %d, create buffer", size)
-	buf := restic.NewBlobBuffer(int(size))
-
-	n, err := r.loadBlob(ctx, id, restic.TreeBlob, buf)
-	if err != nil {
-		return nil, err
-	}
-	buf = buf[:n]
-
 	t := &restic.Tree{}
 	err = json.Unmarshal(buf, t)
 	if err != nil {
@@ -658,11 +747,49 @@ func (r *Repository) SaveTree(ctx context.Context, t *restic.Tree) (restic.ID, e
 	// adds a newline after each object)
 	buf = append(buf, '\n')
 
-	id := restic.Hash(buf)
-	if r.idx.Has(id, restic.TreeBlob) {
-		return id, nil
-	}
-
-	_, err = r.SaveBlob(ctx, restic.TreeBlob, buf, id)
+	id, _, err := r.SaveBlob(ctx, restic.TreeBlob, buf, restic.ID{}, false)
 	return id, err
 }
+
+// Loader allows loading data from a backend.
+type Loader interface {
+	Load(ctx context.Context, h restic.Handle, length int, offset int64, fn func(rd io.Reader) error) error
+}
+
+// DownloadAndHash is all-in-one helper to download content of the file at h to a temporary filesystem location
+// and calculate ID of the contents. Returned (temporary) file is positioned at the beginning of the file;
+// it is reponsibility of the caller to close and delete the file.
+func DownloadAndHash(ctx context.Context, be Loader, h restic.Handle) (tmpfile *os.File, hash restic.ID, size int64, err error) {
+	tmpfile, err = fs.TempFile("", "restic-temp-")
+	if err != nil {
+		return nil, restic.ID{}, -1, errors.Wrap(err, "TempFile")
+	}
+
+	err = be.Load(ctx, h, 0, 0, func(rd io.Reader) (ierr error) {
+		_, ierr = tmpfile.Seek(0, io.SeekStart)
+		if ierr == nil {
+			ierr = tmpfile.Truncate(0)
+		}
+		if ierr != nil {
+			return ierr
+		}
+		hrd := hashing.NewReader(rd, sha256.New())
+		size, ierr = io.Copy(tmpfile, hrd)
+		hash = restic.IDFromHash(hrd.Sum(nil))
+		return ierr
+	})
+	if err != nil {
+		tmpfile.Close()
+		os.Remove(tmpfile.Name())
+		return nil, restic.ID{}, -1, errors.Wrap(err, "Load")
+	}
+
+	_, err = tmpfile.Seek(0, io.SeekStart)
+	if err != nil {
+		tmpfile.Close()
+		os.Remove(tmpfile.Name())
+		return nil, restic.ID{}, -1, errors.Wrap(err, "Seek")
+	}
+
+	return tmpfile, hash, size, err
+}
