@@ -30,14 +30,18 @@ type Snapshot struct {
 // NewSnapshot returns an initialized snapshot struct for the current user and
 // time.
 func NewSnapshot(paths []string, tags []string, hostname string, time time.Time) (*Snapshot, error) {
-	for i, path := range paths {
-		if p, err := filepath.Abs(path); err != nil {
-			paths[i] = p
+	absPaths := make([]string, 0, len(paths))
+	for _, path := range paths {
+		p, err := filepath.Abs(path)
+		if err == nil {
+			absPaths = append(absPaths, p)
+		} else {
+			absPaths = append(absPaths, path)
 		}
 	}
 
 	sn := &Snapshot{
-		Paths:    paths,
+		Paths:    absPaths,
 		Time:     time,
 		Tags:     tags,
 		Hostname: hostname,
@@ -64,15 +68,21 @@ func LoadSnapshot(ctx context.Context, repo Repository, id ID) (*Snapshot, error
 
 // LoadAllSnapshots returns a list of all snapshots in the repo.
 func LoadAllSnapshots(ctx context.Context, repo Repository) (snapshots []*Snapshot, err error) {
-	for id := range repo.List(ctx, SnapshotFile) {
+	err = repo.List(ctx, SnapshotFile, func(id ID, size int64) error {
 		sn, err := LoadSnapshot(ctx, repo, id)
 		if err != nil {
-			return nil, err
+			return err
 		}
 
 		snapshots = append(snapshots, sn)
+		return nil
+	})
+
+	if err != nil {
+		return nil, err
 	}
-	return
+
+	return snapshots, nil
 }
 
 func (sn Snapshot) String() string {
@@ -152,8 +162,10 @@ func (sn *Snapshot) HasTags(l []string) bool {
 	return true
 }
 
-// HasTagList returns true if the snapshot satisfies at least one TagList,
-// so there is a TagList in l for which all tags are included in sn.
+// HasTagList returns true if either
+// - the snapshot satisfies at least one TagList, so there is a TagList in l
+//   for which all tags are included in sn, or
+// - l is empty
 func (sn *Snapshot) HasTagList(l []TagList) bool {
 	debug.Log("testing snapshot with tags %v against list: %v", sn.Tags, l)
 
@@ -191,6 +203,23 @@ func (sn *Snapshot) HasPaths(paths []string) bool {
 	return true
 }
 
+// HasHostname returns true if either
+// - the snapshot hostname is in the list of the given hostnames, or
+// - the list of given hostnames is empty
+func (sn *Snapshot) HasHostname(hostnames []string) bool {
+	if len(hostnames) == 0 {
+		return true
+	}
+
+	for _, hostname := range hostnames {
+		if sn.Hostname == hostname {
+			return true
+		}
+	}
+
+	return false
+}
+
 // Snapshots is a list of snapshots.
 type Snapshots []*Snapshot
 
