@@ -4,10 +4,9 @@ package termstatus
 
 import (
 	"io"
+	"os"
 
-	"golang.org/x/sys/unix"
-
-	isatty "github.com/mattn/go-isatty"
+	"golang.org/x/crypto/ssh/terminal"
 )
 
 // clearCurrentLine removes all characters from the current line and resets the
@@ -24,15 +23,13 @@ func moveCursorUp(wr io.Writer, fd uintptr) func(io.Writer, uintptr, int) {
 // canUpdateStatus returns true if status lines can be printed, the process
 // output is not redirected to a file or pipe.
 func canUpdateStatus(fd uintptr) bool {
-	return isatty.IsTerminal(fd)
-}
-
-// getTermSize returns the dimensions of the given terminal.
-// the code is taken from "golang.org/x/crypto/ssh/terminal"
-func getTermSize(fd uintptr) (width, height int, err error) {
-	ws, err := unix.IoctlGetWinsize(int(fd), unix.TIOCGWINSZ)
-	if err != nil {
-		return -1, -1, err
+	if !terminal.IsTerminal(int(fd)) {
+		return false
 	}
-	return int(ws.Col), int(ws.Row), nil
+	term := os.Getenv("TERM")
+	if term == "" {
+		return false
+	}
+	// TODO actually read termcap db and detect if terminal supports what we need
+	return term != "dumb"
 }
